import copy
import datetime
import json
from pathlib import Path
from typing import Dict, Tuple

import numpy as np
import pandas as pd
import pytest
import zarr
from fastapi.testclient import TestClient

import ultrack
from ultrack import MainConfig, link, segment, solve, to_tracks_layer, tracks_to_zarr
from ultrack.api.app import app
from ultrack.api.database import Experiment, ExperimentStatus
from ultrack.api.settings import settings
from ultrack.api.utils.fs import open_image
from ultrack.api.utils.zarr import get_channels_from_ome_zarr
from ultrack.core.database import clear_all_data
from ultrack.imgproc import detect_foreground, robust_invert
from ultrack.utils import labels_to_contours
from ultrack.utils.array import array_apply
from ultrack.utils.cuda import on_gpu


def _get_ultrack_solution(
    config: MainConfig, foreground: np.ndarray, edges: np.ndarray
) -> Tuple[pd.DataFrame, Dict, zarr.Array]:
    clear_all_data(settings.ultrack_data_config.database_path)

    segment(foreground=foreground, contours=edges, config=config)
    link(config)
    solve(config)

    tracks_df, graph = to_tracks_layer(config)
    segments = tracks_to_zarr(
        config,
        tracks_df,
        store_or_path=zarr.storage.MemoryStore(),
        overwrite=True,
    )

    return tracks_df, graph, segments


@pytest.fixture
def ome_zarr_dataset_path(
    tmp_path: Path, timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array]
) -> str:
    from ome_zarr.io import parse_url
    from ome_zarr.writer import write_image

    path = str(tmp_path / "data.ome.zarr")
    detection, edges, labels = timelapse_mock_data
    data = np.stack([detection, detection, edges, labels], axis=1)

    store = parse_url(path, mode="w").store
    root = zarr.group(store=store)
    write_image(
        image=data, group=root, storage_options=dict(chunks=(1, *detection.chunks))
    )
    root.attrs["ome"] = {
        **root.attrs["ome"],
        "omero": {
            "channels": [
                {"label": v} for v in ["image", "detection", "edges", "labels"]
            ]
        },
    }

    return path


@pytest.fixture
def experiment_instance(
    ome_zarr_dataset_path: str,
    config_instance: MainConfig,
    tmp_path: Path,
) -> Experiment:
    settings.api_results_path = tmp_path

    experiment = Experiment(
        name="PyTest",
        config=config_instance.model_dump(by_alias=True),
        data_url=ome_zarr_dataset_path,
        image_channel_or_path="image",
        edges_channel_or_path="edges",
        detection_channel_or_path="detection",
        labels_channel_or_path="labels",
    )

    return experiment


def test_read_main():
    client = TestClient(app)
    response = client.get("/")
    assert response.status_code == 200
    assert response.json() == {"ultrack.__version__": ultrack.__version__}


def test_config():
    client = TestClient(app)
    response = client.get("/config/default")
    assert response.status_code == 200

    default_config = MainConfig()
    default_config.data_config = None

    assert response.json() == default_config.model_dump(by_alias=True)


def test_manual_segment(experiment_instance: Experiment):
    experiment = copy.deepcopy(experiment_instance)
    # compare the results with the ones obtained from the ultrack module
    # must be done before processing
    node = open_image(experiment.data_url)
    named_data = get_channels_from_ome_zarr(
        node,
        valid_channels=[
            ("detection", experiment.detection_channel_or_path),
            ("edges", experiment.edges_channel_or_path),
        ],
    )
    detection = named_data[experiment.detection_channel_or_path]
    edges = named_data[experiment.edges_channel_or_path]

    tracks_df, graph, segments = _get_ultrack_solution(
        experiment.get_config(), detection, edges
    )

    # processing
    client = TestClient(app)
    with client.websocket_connect("/segment/manual") as websocket:
        json_exp = json.loads(
            json.dumps(
                experiment.model_dump(),
                default=lambda o: (
                    o.isoformat()
                    if isinstance(o, (datetime.date, datetime.datetime))
                    else None
                ),
            )
        )
        websocket.send_json({"experiment": json_exp})
        while experiment.status != ExperimentStatus.SUCCESS:
            response = websocket.receive_json()
            experiment = Experiment.model_validate(response)
            assert experiment.status != ExperimentStatus.ERROR

    tracks_df_api, graph_api = to_tracks_layer(experiment.get_config())
    segments_api = zarr.open(experiment.final_segments_url)

    assert tracks_df.equals(tracks_df_api)
    assert graph == graph_api
    assert np.array_equal(segments_api, segments)


@pytest.mark.parametrize(
    "detect_foreground_kwargs, robust_invert_kwargs",
    [(None, None), ({"sigma": 5}, {"voxel_size": [2, 2, 2]})],
)
def test_auto_detect(
    experiment_instance: Experiment,
    detect_foreground_kwargs: dict,
    robust_invert_kwargs: dict,
):
    experiment = copy.deepcopy(experiment_instance)

    # compare the results with the ones obtained from the ultrack module
    # must be done before processing
    node = open_image(experiment.data_url)
    named_data = get_channels_from_ome_zarr(
        node,
        valid_channels=[
            ("image", experiment.image_channel_or_path),
            ("edges", experiment.edges_channel_or_path),
            ("detection", experiment.detection_channel_or_path),
        ],
    )
    image_data = named_data[experiment.image_channel_or_path]

    # processing
    client = TestClient(app)
    with client.websocket_connect("/segment/auto_detect") as websocket:
        json_exp = json.loads(
            json.dumps(
                experiment.model_dump(),
                default=lambda o: (
                    o.isoformat()
                    if isinstance(o, (datetime.date, datetime.datetime))
                    else None
                ),
            )
        )

        json_request = {"experiment": json_exp}

        if detect_foreground_kwargs:
            json_request["detect_foreground_kwargs"] = detect_foreground_kwargs
        else:
            detect_foreground_kwargs = {}

        if robust_invert_kwargs:
            json_request["robust_invert_kwargs"] = robust_invert_kwargs
        else:
            robust_invert_kwargs = {}

        websocket.send_json(json_request)
        while experiment.status != ExperimentStatus.SUCCESS:
            response = websocket.receive_json()
            experiment = Experiment.model_validate(response)
            assert experiment.status != ExperimentStatus.ERROR

    detection = np.zeros_like(image_data, dtype=float)
    array_apply(
        image_data,
        out_array=detection,
        func=on_gpu(detect_foreground),
        **detect_foreground_kwargs,
    )

    edges = np.zeros_like(image_data, dtype=float)
    array_apply(
        image_data.astype(np.float32),
        out_array=edges,
        func=on_gpu(robust_invert),
        **robust_invert_kwargs,
    )
    tracks_df, graph, segments = _get_ultrack_solution(
        experiment.get_config(), detection, edges
    )

    tracks_df_api, graph_api = to_tracks_layer(experiment.get_config())
    segments_api = zarr.open(experiment.final_segments_url)

    assert tracks_df.equals(tracks_df_api)
    assert graph == graph_api
    assert np.array_equal(segments_api, segments)


@pytest.mark.parametrize(
    "label_to_edges_kwargs",
    [None, {"sigma": 5}],
)
def test_from_labels(experiment_instance: Experiment, label_to_edges_kwargs: dict):
    experiment = copy.deepcopy(experiment_instance)

    # must be loaded before processing
    node = open_image(experiment.data_url)
    named_data = get_channels_from_ome_zarr(
        node, valid_channels=[("labels", experiment.labels_channel_or_path)]
    )
    label_data = named_data[experiment.labels_channel_or_path]

    # processing
    client = TestClient(app)
    with client.websocket_connect("/segment/labels") as websocket:
        json_exp = json.loads(
            json.dumps(
                experiment.model_dump(),
                default=lambda o: (
                    o.isoformat()
                    if isinstance(o, (datetime.date, datetime.datetime))
                    else o
                ),
            )
        )
        if label_to_edges_kwargs:
            websocket.send_json(
                {"experiment": json_exp, "label_to_edges_kwargs": label_to_edges_kwargs}
            )
        else:
            websocket.send_json({"experiment": json_exp})
            label_to_edges_kwargs = {}
        while experiment.status != ExperimentStatus.SUCCESS:
            response = websocket.receive_json()
            experiment = Experiment.model_validate(response)
            assert experiment.status != ExperimentStatus.ERROR

    # compare the results with the ones obtained from the ultrack module
    detection, edges = labels_to_contours(
        label_data,
        foreground_store_or_path=zarr.storage.MemoryStore(),
        contours_store_or_path=zarr.storage.MemoryStore(),
        **label_to_edges_kwargs,
    )

    tracks_df, graph, segments = _get_ultrack_solution(
        experiment.get_config(), detection, edges
    )

    tracks_df_api, graph_api = to_tracks_layer(experiment.get_config())
    segments_api = zarr.open(experiment.final_segments_url)

    assert tracks_df.equals(tracks_df_api)
    assert graph == graph_api
    assert np.array_equal(segments_api, segments)


def test_output_experiment(experiment_instance: Experiment):
    client = TestClient(app)
    with client.websocket_connect("/segment/auto_detect") as websocket:
        json_exp = json.loads(
            json.dumps(
                experiment_instance.model_dump(),
                default=lambda o: (
                    o.isoformat()
                    if isinstance(o, (datetime.date, datetime.datetime))
                    else None
                ),
            )
        )
        websocket.send_json({"experiment": json_exp})

        # skip the first message
        websocket.receive_json()

        while experiment_instance.status != ExperimentStatus.SUCCESS:
            response = websocket.receive_json()
            experiment_instance = Experiment.model_validate(response)
            assert experiment_instance.status != ExperimentStatus.ERROR

        assert "robust_invert" in experiment_instance.err_log


def test_available_configs(experiment_instance: Experiment):
    client = TestClient(app)

    response = client.get("/config/available")
    assert response.status_code == 200

    for _, data in response.json().items():
        experiment = copy.deepcopy(experiment_instance)
        uri = data["link"]
        config_metadata = data["config"]
        config_metadata["experiment"]["data_url"] = experiment.data_url
        config_metadata["experiment"]["name"] = "PyTest"
        config_metadata["experiment"][
            "image_channel_or_path"
        ] = experiment.image_channel_or_path
        config_metadata["experiment"][
            "edges_channel_or_path"
        ] = experiment.edges_channel_or_path
        config_metadata["experiment"][
            "detection_channel_or_path"
        ] = experiment.detection_channel_or_path
        config_metadata["experiment"][
            "labels_channel_or_path"
        ] = experiment.labels_channel_or_path

        with client.websocket_connect(uri) as websocket:
            json_exp = json.loads(
                json.dumps(
                    config_metadata["experiment"],
                    default=lambda o: (
                        o.isoformat()
                        if isinstance(o, (datetime.date, datetime.datetime))
                        else None
                    ),
                )
            )
            websocket.send_json({"experiment": json_exp})

            # skip the first message
            websocket.receive_json()

            while experiment.status != ExperimentStatus.SUCCESS:
                response = websocket.receive_json()
                experiment = Experiment.model_validate(response)
                assert experiment.status != ExperimentStatus.ERROR


# def test_exception_logging(experiment_instance: Experiment):
#     experiment_instance.labels_channel = "non_existent_channel"
#
#     client = TestClient(app)
#     with client.websocket_connect("/segment/labels") as websocket:
#         json_exp = json.loads(
#             json.dumps(
#                 experiment_instance.dict(),
#                 default=lambda o: o.isoformat()
#                 if isinstance(o, (datetime.date, datetime.datetime))
#                 else None,
#             )
#         )
#         websocket.send_json({"experiment": json_exp})
#
#         # skip the first message
#         websocket.receive_json()
#
#         while experiment_instance.status != ExperimentStatus.ERROR:
#             response = websocket.receive_json()
#             experiment_instance = Experiment.model_validate(response)
#
#         assert "Exception" in experiment_instance.err_log
#         assert "Traceback" in experiment_instance.err_log
#         assert "non_existent_channel" in experiment_instance.err_log
#         assert "ValueError" in experiment_instance.err_log
#         assert experiment_instance.status == ExperimentStatus.ERROR


# def test_interruption(experiment_instance: Experiment):
#     client = TestClient(app)
#     with client.websocket_connect("/segment/manual") as websocket:
#         json_exp = json.loads(
#             json.dumps(
#                 experiment_instance.dict(),
#                 default=lambda o: o.isoformat()
#                 if isinstance(o, (datetime.date, datetime.datetime))
#                 else None,
#             )
#         )
#         websocket.send_json({"experiment": json_exp})
#         response = websocket.receive_json()
#         experiment_instance = Experiment.model_validate(response)
#         response = client.post(f"/stop/{experiment_instance.id}")
#         assert response.status_code == 200
#         response = websocket.receive_json()
#         experiment_instance = Experiment.model_validate(response)
#         assert experiment_instance.status == ExperimentStatus.ERROR



from typing import Dict, List

import numpy as np
import pandas as pd
from numba import typed, types

from ultrack.tracks import (
    left_first_search,
    sort_track_ids,
    sort_trees_by_length,
    sort_trees_by_max_radius,
)
from ultrack.utils.constants import NO_PARENT


def test_sortrees_by_length() -> None:
    """
                 5
                 --
            2   /
         -----
     1  /      \\
    ---          -----
       \\         6
         --
         3

        --------
        4
    """
    df = pd.DataFrame(
        {
            "track_id": [
                1,
                1,
                1,
                2,
                2,
                2,
                2,
                2,
                3,
                3,
                4,
                4,
                4,
                4,
                4,
                4,
                4,
                4,
                5,
                5,
                6,
                6,
                6,
                6,
                6,
            ],
        }
    )
    graph = {6: 2, 5: 2, 2: 1, 3: 1}

    sorted = sort_trees_by_length(df, graph)

    assert len(sorted) == 2

    for track_id in [1, 2, 3, 5, 6]:
        assert track_id in sorted[0]["track_id"]
        assert track_id not in sorted[1]["track_id"]

    assert not np.any(4 == sorted[0]["track_id"])
    assert np.all(4 == sorted[1]["track_id"])

    reconstr_df = pd.concat(sorted).sort_values("track_id")
    assert np.allclose(reconstr_df, df)


def _to_numba_dict(
    graph: Dict[int, List[int]],
) -> types.DictType(types.int64, types.ListType(types.int64)):
    """Convert a Python dictionary to a Numba dictionary."""

    nb_graph = typed.Dict.empty(types.int64, types.ListType(types.int64))
    for k, v in graph.items():
        children = typed.List.empty_list(types.int64)
        for child in v:
            children.append(child)
        nb_graph[k] = children

    return nb_graph


def test_left_first_search_single_node() -> None:
    # Test for a single node binary tree
    graph = _to_numba_dict({})
    result = left_first_search(1, graph)
    assert result == [1]


def test_left_first_search_complete_binary_tree() -> None:
    # Test for a complete binary tree (all nodes have two children)
    graph = _to_numba_dict({1: [2, 3], 2: [4, 5], 3: [6, 7]})
    result = left_first_search(1, graph)
    assert result == [4, 2, 5, 1, 6, 3, 7]


def test_sort_track_ids_forest() -> None:
    # Test for multiple binary trees forming a forest
    data = {
        "track_id": np.repeat([1, 2, 3, 4, 5, 6, 7], 3),
        "parent_track_id": np.repeat([NO_PARENT, 1, 1, NO_PARENT, 4, 4, NO_PARENT], 3),
    }

    tracks_df = pd.DataFrame(data)
    result = sort_track_ids(tracks_df)
    assert np.array_equal(result, np.array([2, 1, 3, 5, 4, 6, 7]))


def test_sort_trees_by_max_radius() -> None:
    tracks_df = pd.DataFrame(
        {
            "track_id": [1, 1, 2, 3, 4, 5, 6, 7],
            "parent_track_id": [NO_PARENT, NO_PARENT, NO_PARENT, 2, 2, NO_PARENT, 5, 5],
            "t": [0, 1, 0, 1, 1, 0, 1, 1],
            "x": [1, 2, 0, 1, 3, 7, 1, 2],
        }
    )

    expected_track_ids = [[2, 3, 4], [5, 6, 7], [1, 1]]

    sorted_tracks = sort_trees_by_max_radius(tracks_df, scale=1.0)

    for track_ids, expected_ids in zip(sorted_tracks, expected_track_ids):
        assert np.array_equal(track_ids["track_id"].to_numpy(), expected_ids)

    # no radius
    sorted_tracks = sort_trees_by_max_radius(tracks_df)

    for track_ids, expected_ids in zip(sorted_tracks, expected_track_ids):
        assert np.array_equal(track_ids["track_id"].to_numpy(), expected_ids)



import random
import shutil
import warnings
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, Generator, Literal, Optional, Sequence, Union

import dask.array as da
import napari
import numpy as np
import pandas as pd
from napari.components import Camera
from napari.layers import Layer
from numpy.typing import ArrayLike
from pandas.core.groupby import DataFrameGroupBy
from tqdm import tqdm

from ultrack.tracks import sort_trees_by_length, split_tracks_df_by_lineage

try:
    from napari_animation import Animation

except (ImportError, ModuleNotFoundError):
    warnings.warn(
        "To export tracks as videos, please install napari-animation: pip install napari-animation"
    )


def _disable_thumbnails(viewer: napari.Viewer) -> None:
    for l in viewer.layers:
        l._update_thumbnail = lambda *args, **kwargs: None


@contextmanager
def _block_refresh(viewer: napari.Viewer) -> Generator[None, None, None]:
    """
    Disable layer refresh during context block.
    Useful to speed up state update when creating animation

    Parameters
    ----------
    viewer : napari.Viewer
        Napari viewer.
    """
    prev_refresh = [l.refresh for l in viewer.layers]

    for l in viewer.layers:
        l.refresh = lambda *args, **kwargs: None

    yield

    for l, v in zip(viewer.layers, prev_refresh):
        l.refresh = v

    # NOTE: after napari 0.5.0 current solution can be replaced with:
    # prev_values = [
    #     l._refresh_blocked
    #     for l in viewer.layers
    # ]

    # for l in viewer.layers:
    #     l._refresh_blocked = True

    # yield

    # for l, v in zip(viewer.layers, prev_values):
    #     l._refresh_blocked = v


def _optimize_layers(viewer: napari.Viewer) -> None:
    """
    Convert layers to data types used by vispy.
    """
    size_threshold = 2048 * 2048 * 512
    for l in viewer.layers:
        if isinstance(l.data, da.Array) or l.data.size <= size_threshold:
            try:
                if l.data.dtype == np.uint16 or l.data.dtype == np.int16:
                    l.data = l.data.astype(np.int32)
                elif l.data.dtype == np.float16:
                    l.data = l.data.astype(np.float32)
            except Exception:
                pass


def record_tracks(
    viewer: napari.Viewer,
    tracks_df: pd.DataFrame,
    output_directory: Path,
    overwrite: bool,
    orthogonal_proj: bool,
    time_step: int,
    tracks_layer_kwargs: Dict[str, Any],
    tracks_layer_attrs: Dict[str, Any],
) -> None:
    """
    Record tracks as individual videos.

    Parameters
    ----------
    viewer : napari.Viewer
        Napari viewer with additional layers and configuration already setup.
    tracks_df : pd.DataFrame
        Tracks dataframe.
    output_directory : Path
        Output directory.
    overwrite : bool
        Whether to overwrite existing videos, otherwise they're skipped.
    orthogonal_proj : bool
        Whether to record orthogonal projections of entire tracks_df.
    time_step : int
        Napari animation keyframe time step.
    tracks_layer_kwargs : Dict[str, Any]
        Keyword arguments for the tracks layer creation.
    tracks_layer_attrs : Dict[str, Any]
        Keyword attributes set after the tracks layer is created.
    """
    root_id = int(tracks_df["track_id"].iloc[0])

    output_directory = output_directory / str(root_id)
    if output_directory.exists() and overwrite:
        shutil.rmtree(output_directory)

    output_directory.mkdir(exist_ok=True, parents=True)

    if "z" not in tracks_df.columns:
        cols = ["t", "y", "x"]
        ncols = 3
    else:
        cols = ["t", "z", "y", "x"]
        ncols = 4

    track_update_freq = 5

    # previous setup
    order = viewer.dims.order
    zoom = viewer.camera.zoom

    # ----- Single tracklet video ----- #
    viewer.dims.ndisplay = 2

    for axis_order in (0, 1):

        # converting to zx-slice
        if axis_order == 1:
            if ncols == 3:
                print("Skipping zx-slicing for 2D data.")
                continue
            new_order = order[:-3] + (order[-2], order[-3], order[-1])
            viewer.dims.order = new_order
            center_idx = [-3, -1]
        else:
            center_idx = [-2, -1]

        viewer.camera.zoom = zoom

        for track_id, tracklet in tracks_df.groupby("track_id", sort=True):
            out_path = output_directory / f"{track_id}_axis_{axis_order}_roi.mp4"
            if out_path.exists():
                print(f"Skipping {out_path}, it already exists.")
                continue

            tracks_layer = viewer.add_tracks(
                tracklet[["track_id", *cols]],
                **tracks_layer_kwargs,
            )
            for k, v in tracks_layer_attrs.items():
                setattr(tracks_layer, k, v)

            with _block_refresh(viewer):

                animation = Animation(viewer)

                for i, (_, s_track) in enumerate(
                    tracklet.groupby("t", sort=True, as_index=False)
                ):
                    if i % track_update_freq != 0:
                        continue

                    pos = s_track[cols].iloc[0].to_numpy()  # t, (z), y, x
                    viewer.dims.set_point(range(ncols), pos)
                    viewer.camera.center = pos[center_idx]
                    animation.capture_keyframe(track_update_freq * time_step)

                pos = s_track[cols].iloc[0].to_numpy()  # t, (z), y, x
                viewer.dims.set_point(range(ncols), pos)
                viewer.camera.center = pos[center_idx]
                animation.capture_keyframe((i % track_update_freq) * time_step)

            animation.animate(out_path, fps=60)

            viewer.layers.remove(tracks_layer)

    viewer.dims.order = order
    viewer.camera.zoom = zoom

    if not orthogonal_proj:
        return
    elif ncols == 3:
        print("Skipping orthogonal projections for 2D data.")
        return

    # ----- projection videos ----- #
    tracks_layer = viewer.add_tracks(
        tracks_df[["track_id", *cols]], **tracks_layer_kwargs
    )
    for k, v in tracks_layer_attrs.items():
        setattr(tracks_layer, k, v)

    viewer.dims.ndisplay = 3
    camera_angles = [(0, 0, 90), (0, 0, 180), (-90, 90, 0)]

    for proj_idx, angles in enumerate(camera_angles):

        out_path = output_directory / f"proj_{proj_idx}.mp4"
        if out_path.exists():
            print(f"Skipping {out_path}, it already exists.")
            continue

        t_min, t_max = tracks_df["t"].min(), tracks_df["t"].max()

        with _block_refresh(viewer):

            viewer.dims.set_point(0, t_min)
            viewer.camera.angles = angles

            animation = Animation(viewer)

            animation.capture_keyframe()

            viewer.dims.set_point(0, t_max)

            animation.capture_keyframe(int((t_max - t_min) * time_step))

        animation.animate(out_path, fps=60)

    # resetting
    viewer.layers.remove(tracks_layer)
    viewer.dims.order = order
    viewer.camera.zoom = zoom


def tracks_df_to_videos(
    viewer: napari.Viewer,
    tracks_df: pd.DataFrame,
    output_directory: Union[Path, str],
    num_lineages: Optional[int] = None,
    sort: Literal["random", "length", "id"] = "length",
    overwrite: bool = False,
    orthogonal_proj: bool = True,
    time_step: int = 5,
    tracks_layer_kwargs: Dict[str, Any] = {},
    tracks_layer_attrs: Dict[str, Any] = {},
) -> None:
    """
    Record tracks as individual videos.

    Parameters
    ----------
    viewer : napari.Viewer
        Napari viewer with additional layers and configuration already setup.
    tracks_df : pd.DataFrame
        Tracks dataframe, IMPORTANT: z, y, x should be at world (physical) coordinate system.
    output_directory : Path
        Output directory.
    num_lineages : int
        Number of lineages to record.
    sort : Literal["random", "length", "id"], optional
        Method to sort the lineages, by default "length".
    overwrite : bool, optional
        Whether to overwrite existing videos, otherwise they're skipped.
    orthogonal_proj : bool, optional
        Whether to record orthogonal projections of entire tracks_df.
    time_step : int, optional
        Napari animation keyframe time step.
    tracks_layer_kwargs : Dict[str, Any], optional
        Keyword arguments for the tracks layer creation.
    tracks_layer_attrs : Dict[str, Any], optional
        Keyword attributes set after the tracks layer is created.
    """

    if isinstance(output_directory, str):
        output_directory = Path(output_directory)

    output_directory.mkdir(exist_ok=True, parents=True)

    sort = sort.lower()

    if sort == "length":
        trees = sort_trees_by_length(tracks_df)
    else:
        trees = split_tracks_df_by_lineage(tracks_df)
        if sort == "random":
            random.shuffle(trees)
        elif sort != "id":
            raise ValueError(f"Unknown sort method: {sort}")

    if num_lineages is None:
        num_lineages = len(trees)

    if len(trees) < num_lineages:
        warnings.warn(
            f"Requested {num_lineages} lineages, but only {len(trees)} available."
        )

    _disable_thumbnails(viewer)
    _optimize_layers(viewer)

    for i, tree in tqdm(enumerate(trees), total=num_lineages):
        if i >= num_lineages:
            break

        record_tracks(
            viewer=viewer,
            tracks_df=tree,
            overwrite=overwrite,
            output_directory=output_directory,
            orthogonal_proj=orthogonal_proj,
            time_step=time_step,
            tracks_layer_kwargs=tracks_layer_kwargs,
            tracks_layer_attrs=tracks_layer_attrs,
        )


def _update_center_planes(
    viewer: napari.Viewer,
    clipping_planes_layers: Sequence[Layer],
    box_size: int,
    position: ArrayLike,
) -> None:
    """
    Update clipping planes to center around a position.

    Parameters
    ----------
    viewer : napari.Viewer
        Napari viewer.
    clipping_planes_layers : Sequence[Layer]
        Layers to be updated with clipping planes.
    box_size : int
        Size of the box around the position.
    position : ArrayLike
        3D position to center the box around.
    """

    if not clipping_planes_layers:
        return

    if len(position) != 3:
        raise ValueError(f"Position should be 3D. Found {len(position)} dimensions.")

    position = np.asarray(position)

    half_size = box_size // 2
    planes = []
    for i in range(3):
        for d in (-1, 1):
            p = position.copy()
            p[i] -= d * half_size
            n = np.zeros(3)
            n[i] = d
            planes.append({"normal": n, "position": p})

    n = np.asarray(viewer.camera.view_direction)
    planes.append({"normal": n, "position": position + 1})
    planes.append({"normal": -n, "position": position - 1})

    for l in clipping_planes_layers:
        l.experimental_clipping_planes = planes


def _linear_interp(
    current_t: int,
    start_t: int,
    end_t: Optional[int],
    start_val: ArrayLike,
    end_val: Optional[ArrayLike],
) -> ArrayLike:
    """
    Linear interpolation between two values.

    Parameters
    ----------
    current_t : int
        Current time.
    start_t : int
        Start time.
    end_t : int
        End time.
    start_val : ArrayLike
        Start value.
    end_val : ArrayLike
        End value.

    Returns
    -------
    ArrayLike
        Interpolated value.
    """
    if end_t is None:
        return start_val
    w = (current_t - start_t) / (end_t - start_t)
    w = np.clip(w, 0, 1)
    return start_val * (1 - w) + end_val * w


@dataclass
class CameraAttr:
    prev_t: int
    prev: ArrayLike
    next: Optional[ArrayLike]
    next_t: Optional[int]
    next_camera_attr: Optional["CameraAttr"] = None

    def current(self, t: int) -> ArrayLike:
        return _linear_interp(t, self.prev_t, self.next_t, self.prev, self.next)

    def maybe_next(self, t: int) -> "CameraAttr":
        if self.next_t is None or t < self.next_t:
            return self
        return self.next_camera_attr

    def create_next(self, t: int, val: int) -> "CameraAttr":
        self.next_t = t
        self.next = val
        self.next_camera_attr = CameraAttr(t, val, None, None)
        return self.next_camera_attr


def _maybe_get_auto_value(
    tracks_by_t: DataFrameGroupBy,
    t: int,
    key: str,
    value: ArrayLike,
    origin: ArrayLike,
) -> ArrayLike:
    """
    Get auto value for camera attributes.

    Parameters
    ----------
    tracks_by_t : DataFrameGroupBy
        Tracks dataframe grouped by time.
    t : int
        Time point.
    key : str
        Camera attribute key.
    value : ArrayLike
        Value to check.
    origin : ArrayLike
        Origin position.

    Returns
    -------
    ArrayLike
        Updated value.
    """
    if value == "auto":
        if key == "center":
            value = tracks_by_t.get_group(t)[["z", "y", "x"]].mean().to_numpy()
        elif key == "angles":
            point = tracks_by_t.get_group(t)[["z", "y", "x"]].mean().to_numpy()
            dummy_camera = Camera()
            dummy_camera.set_view_direction(origin - point)
            value = dummy_camera.angles
        else:
            raise ValueError(
                f"Only 'center' and 'angles' can be set to 'auto'. Found in '{key}'."
            )

    return np.asarray(value)


def _setup_camera_attrs(
    tracks_by_t: DataFrameGroupBy,
    cam: Camera,
    t_min: int,
    time_to_camera_kwargs: Dict[int, Dict[str, Any]],
    origin: ArrayLike,
) -> Dict[int, CameraAttr]:
    """
    Setup dictionary of linked lists for camera attributes for each time point.

    Parameters
    ----------
    tracks_by_t : DataFrameGroupBy
        Tracks dataframe grouped by time.
    cam : napari.components.Camera
        Napari camera.
    t_min : int
        Minimum time point.
    time_to_camera_kwargs : Dict[int, Dict[str, Any]]
        Dictionary of time points to camera attributes.
    origin : ArrayLike
        Origin position.

    Returns
    -------
    Dict[int, CameraAttr]
        Dictionary of camera attributes for each time point.
    """
    cam_keys = ("zoom", "angles", "center")
    for t, cam_kwargs in time_to_camera_kwargs.items():
        for k in cam_kwargs:
            if k not in cam_keys:
                raise ValueError(f"Unknown camera attribute '{k}' at key '{t}'.")

    if t_min not in time_to_camera_kwargs:
        time_to_camera_kwargs[t_min] = {}

    cam_attrs: Dict[int, CameraAttr] = {}
    for c in cam_keys:
        value = time_to_camera_kwargs[t_min].get(c, getattr(cam, c))

        value = _maybe_get_auto_value(tracks_by_t, t_min, c, value, origin)

        cam_attrs[c] = CameraAttr(t_min, value, None, None)

    tmp_cam_attrs = cam_attrs.copy()

    sorted_cam_t = list(sorted(time_to_camera_kwargs.keys()))
    for t in sorted_cam_t:
        if t == t_min:
            continue

        for k, v in time_to_camera_kwargs[t].items():
            v = _maybe_get_auto_value(tracks_by_t, t, k, v, origin)
            tmp_cam_attrs[k] = tmp_cam_attrs[k].create_next(t, v)

    return cam_attrs


def tracks_df_to_3D_video(
    viewer: napari.Viewer,
    tracks_df: pd.DataFrame,
    output_path: Union[Path, str],
    keyframe_step: int = 5,
    update_step: int = 5,
    clipping_planes_layers: Sequence[Layer] = (),
    clipping_box_size: int = 25,
    time_to_camera_kwargs: Dict[int, Dict[str, Any]] = {},
    time_to_viewer_cb: Dict[int, Callable[[napari.Viewer], None]] = {},
    origin: Optional[ArrayLike] = None,
    overwrite: bool = False,
) -> None:
    """
    Record 3D video of a single lineage with optional moving camera and clipping planes.

    Parameters
    ----------
    viewer : napari.Viewer
        Napari viewer with additional layers and configuration already setup.
        IMPORTANT: Including the tracks layer.
    tracks_df : pd.DataFrame
        Tracks dataframe, IMPORTANT: z, y, x should be at world (physical) coordinate system.
    output_path : Path
        Output path for the video.
    keyframe_step : int
        Keyframe step for napari animation, greater value creates longer movies.
    update_step : int
        Update time step, how often to update the viewer.
        The greater the values more often the positions will be updated.
    clipping_planes_layers : Sequence[Layer]
        Layers to be updated with clipping planes.
    clipping_box_size : int
        Size of the clipping plane bounding box around the position in world coordinates.
    time_to_camera_kwargs : Dict[int, Dict[str, Any]]
        Dictionary of time points to update camera attributes.
        Can include 'zoom', 'angles', 'center' per time point, for example:
        {0: {"zoom": 1.5, "angles": (0, 0, 90), "center": (0, 0, 0)}}
        "center" supports "auto" to set the center to the mean of the tracks at that time point.
    time_to_viewer_cb : Dict[int, Callable[[napari.Viewer]]]
        Dictionary of time points to viewer callback functions.
        Callbacks are called at the provided time point.
        Useful to update viewer settings, for example:
            ``{0: lambda viewer: viewer.dims.ndisplay = 3}``
    origin : ArrayLike
        Volume origin position, used to automatically set camera angles.
        When not provided, the mean of the tracks is used.
    overwrite : bool
        Whether to overwrite existing video.
    """

    output_path = Path(output_path)
    if output_path.exists() and not overwrite:
        raise FileExistsError(f"Output path '{output_path}' already exists.")

    if not output_path.name.endswith(".mp4"):
        raise ValueError(f"Output path should be a .mp4 file. Found '{output_path}'.")

    tracks_df["t"] = tracks_df["t"].astype(int)
    t_min = tracks_df["t"].min().item()

    cols = ["z", "y", "x"]
    crange = tuple(range(1, 4))

    cam = viewer.camera
    tracks_by_t = tracks_df.groupby("t")

    if origin is None:
        origin = tracks_df[cols].mean().to_numpy()

    cam_attrs = _setup_camera_attrs(
        tracks_by_t,
        cam,
        t_min,
        time_to_camera_kwargs,
        origin=origin,
    )

    _disable_thumbnails(viewer)

    with _block_refresh(viewer):
        _optimize_layers(viewer)

        animation = Animation(viewer)

        for t in range(t_min, int(tracks_df["t"].max().item()) + 1):

            if t in time_to_viewer_cb:
                time_to_viewer_cb[t](viewer)

            if t != t_min and t % update_step != 0:
                continue

            for k, c in list(cam_attrs.items()):
                value = c.current(t)
                setattr(cam, k, value)
                cam_attrs[k] = c.maybe_next(t)

            center = tracks_by_t.get_group(t)[cols].mean().to_numpy()

            viewer.dims.set_point(0, t)
            viewer.dims.set_point(crange, center)
            _update_center_planes(
                viewer,
                clipping_planes_layers,
                clipping_box_size,
                center,
            )

            animation.capture_keyframe(update_step * keyframe_step)

        animation.capture_keyframe((t % update_step) * keyframe_step)

    animation.animate(output_path, fps=60)


def tracks_df_to_moving_2D_plane_video(
    viewer: napari.Viewer,
    tracks_df: pd.DataFrame,
    output_path: Union[Path, str],
    plane_mov_scale: float = 1.0,
    keyframe_step: int = 5,
    tracks_layer_kwargs: Dict[str, Any] = {},
    overwrite: bool = False,
) -> None:
    """
    Record 2D video of a single lineage with moving z-plane.

    Parameters
    ----------
    viewer : napari.Viewer
        Napari viewer with additional layers and configuration already setup.
    tracks_df : pd.DataFrame
        Tracks dataframe.
    output_path : Path
        Output path for the video.
    plane_mov_scale : float
        Scale factor for the moving image scale.
    keyframe_step : int
        Keyframe step for napari animation, greater value creates longer movies.
    tracks_layer_kwargs : Dict[str, Any]
        Keyword arguments for the tracks layer creation.
    overwrite : bool
        Whether to overwrite existing video.
    """
    output_path = Path(output_path)
    if output_path.exists() and not overwrite:
        raise FileExistsError(f"Output path '{output_path}' already exists.")

    tracks_df["z"] = tracks_df["t"] * plane_mov_scale

    if viewer.dims.ndisplay != 3:
        warnings.warn(
            "Setting viewer to 3D display. Set viewer.dims.ndisplay = 3 and select desired camera settings."
        )

    other_layers = list(viewer.layers)

    for l in other_layers:
        # expanding z
        scale = list(l.scale)
        scale.insert(1, plane_mov_scale)
        l.data = l.data[:, None]
        l.scale = scale

    viewer.add_tracks(
        tracks_df[["track_id", "t", "z", "y", "x"]],
        **tracks_layer_kwargs,
    )

    _disable_thumbnails(viewer)

    viewer.layers.move(-1, 0)

    with _block_refresh(viewer):
        _optimize_layers(viewer)

        animation = Animation(viewer)
        t_min, t_max = tracks_df["t"].min(), tracks_df["t"].max()

        viewer.dims.set_point(0, t_min)
        animation.capture_keyframe()

        delta_t = t_max - t_min
        viewer.dims.set_point(0, t_max)
        for l in other_layers:
            T = list(l.translate)
            T[1] += delta_t * plane_mov_scale
            l.translate = T

        animation.capture_keyframe(delta_t * keyframe_step)

    animation.animate(output_path, fps=60)



import pandas as pd
import pytest

from ultrack.core.export import tracks_layer_to_networkx
from ultrack.utils.constants import NO_PARENT


@pytest.mark.parametrize("children_to_parent", [True, False])
def test_to_networkx(
    children_to_parent: bool,
) -> None:
    tracks_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "parent_id": [NO_PARENT, 1, 2, 2],
            "track_id": [1, 1, 2, 3],
            "t": [0, 1, 2, 2],
            "z": [0, 0, 0, 0],
            "y": [10, 20, 30, 10],
            "x": [1, 2, 3, 1],
        }
    )
    tracks_df.set_index("id", drop=False, inplace=True)

    graph = tracks_layer_to_networkx(tracks_df, children_to_parent=children_to_parent)

    assert len(graph.nodes) == tracks_df.shape[0]
    assert len(graph.edges) == tracks_df.shape[0] - 1

    for node, data in graph.nodes(data=True):
        assert node in tracks_df.index
        for key, value in data.items():
            assert value == tracks_df.loc[node, key]

    if children_to_parent:
        for edge in [(2, 1), (3, 2), (4, 2)]:
            assert edge in graph.edges
    else:
        for edge in [(1, 2), (2, 3), (2, 4)]:
            assert edge in graph.edges


def test_to_networkx_without_parent_id() -> None:
    tracks_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "track_id": [1, 1, 2, 3],
            "t": [0, 1, 2, 2],
            "z": [0, 0, 0, 0],
            "y": [10, 20, 30, 10],
            "x": [1, 2, 3, 1],
        }
    )
    tracks_df.set_index("id", drop=False, inplace=True)

    graph = tracks_layer_to_networkx(tracks_df)

    assert len(graph.nodes) == tracks_df.shape[0]
    assert len(graph.edges) == 1

    for node in graph.nodes:
        assert node in tracks_df.index

    assert (1, 2) in graph.edges



from typing import Dict, List, Tuple, Union

import ome_zarr
import zarr


def _get_omero_spec(node: ome_zarr.reader.Node) -> Dict:
    """Get the OMERO spec from the ome-zarr image.

    Parameters
    ----------
    node : ome_zarr.reader.Node
        The ome-zarr image node.

    Raises
    ------
    ValueError
        If the image doesn't have an OMERO spec.
    """
    for spec in node.specs:
        if isinstance(spec, ome_zarr.reader.OMERO):
            return spec.image_data["channels"]
    raise ValueError("No OMERO spec found.")


def get_channels_from_ome_zarr(
    node: ome_zarr.reader.Node, valid_channels: List[Tuple[str, str]]
) -> Dict[str, Dict[str, Union[zarr.Array, dict]]]:
    """Get the data of the specified channels from the ome-zarr image.

    Parameters
    ----------
    node : ome_zarr.reader.Node
        The ome-zarr image node.
    valid_channels : List[Tuple[str, str]]
        List of tuples with the name of the channel and the channel name in the image.

    Raises
    ------
    ValueError
        If the image doesn't have any of the specified channels.

    Returns
    -------
    Dict[str, Dict[str, zarr.Array]]
        Dictionary with the data of the specified channels.

    Examples
    --------
    >>> import ultrack.api.utils as api_utils
    >>> api_utils.fs.open_image("https://public.czbiohub.org/royerlab/zebrahub/imaging/single-objective/ZSNS001.ome.zarr") # noqa
    >>> named_data = api_utils.zarr.get_channels_from_ome_zarr(node, [("image", "fused")])
    >>> zarr_data = named_data["image"]["data"]
    """
    valid_channels = {
        name: channel for name, channel in valid_channels if channel is not None
    }

    if len(valid_channels) == 0:
        raise ValueError(f"node {node} doesn't have any channels.")

    channel_dim_index = -1

    for i, axis in enumerate(node.metadata["axes"]):
        if axis["type"] == "channel":
            channel_dim_index = i
            break

    if channel_dim_index == -1:
        raise ValueError("The image doesn't have a channel axis.")

    named_data = {}
    for name, channel in valid_channels.items():
        try:
            image_index = node.metadata["channel_names"].index(channel)
        except ValueError as ex:
            raise ValueError(
                f"node {node} doesn't have a channel named {channel}. {ex}"
            )

        channel_slice = [
            slice(None) if axis_idx != channel_dim_index else image_index
            for axis_idx in range(node.data[0].ndim)
        ]
        image_data = node.data[0][tuple(channel_slice)]
        named_data[name] = image_data

    return named_data



from typing import Callable, List

import napari
import numpy as np
import pytest

from ultrack.widgets import LabelsToContoursWidget

pytest.skip(reason="DEPRECATED", allow_module_level=True)


def test_labels_to_contours_widget(
    make_napari_viewer: Callable[[], napari.Viewer],
    multiple_labels: List[np.ndarray],  # noqa: F811
) -> None:

    viewer = make_napari_viewer()
    for label in multiple_labels:
        viewer.add_labels(label)

    viewer.layers.selection.update(viewer.layers)

    assert len(viewer.layers.selection) == len(viewer.layers)

    widget = LabelsToContoursWidget(viewer)
    widget._run_btn.clicked.emit()

    assert "contours" in viewer.layers

    foreground = viewer.layers["foreground"].data

    for lb in multiple_labels:
        for t in range(foreground.shape[0]):
            mask = lb[t] > 0
            assert np.all(foreground[t][mask] > 0)



import logging
from typing import Dict, List, Optional, Sequence
from warnings import warn

import napari
import numpy as np
import pandas as pd
import sqlalchemy as sqla
from magicgui.widgets import CheckBox, FloatSlider, PushButton
from napari.layers import Labels
from sqlalchemy.orm import Session

from ultrack.core.database import LinkDB, NodeDB
from ultrack.core.segmentation.node import Node
from ultrack.widgets._generic_data_widget import GenericDataWidget

logging.basicConfig()
logging.getLogger("sqlachemy.engine").setLevel(logging.INFO)

LOG = logging.getLogger(__name__)


class HypothesesVizWidget(GenericDataWidget):
    def __init__(self, viewer: napari.Viewer) -> None:
        super().__init__(viewer, "Hypotheses Viz.")
        self._segm_layer_name = "Segm. Hypotheses"
        self._link_layer_name = "Links"
        self._nodes: Dict[int, Node] = {}
        self._hier_ids: List[int] = []

        self._load_btn = PushButton(text="Load Segm.")
        self._load_btn.changed.connect(self._on_load_segm)
        self.append(self._load_btn)

        self._area_threshold_w = FloatSlider(label="Area threshold", min=0, max=0)
        self._area_threshold_w.changed.connect(self._on_threshold_update)
        self.append(self._area_threshold_w)

        self._link_w = CheckBox(text="Show links", value=False)
        self.append(self._link_w)

    def _on_config_changed(self) -> None:
        self._ndim = len(self._shape)

    @property
    def _shape(self) -> Sequence[int]:
        return self.config.metadata.get("shape", [])

    def _on_load_segm(self) -> None:
        time = self._time
        engine = sqla.create_engine(self.config.database_path)
        with Session(engine) as session:
            query = (
                session.query(NodeDB.pickle, NodeDB.t_hier_id)
                .where(NodeDB.t == time)
                .order_by(NodeDB.area)
            )
            self._nodes, self._hier_ids = zip(*query)
            # overlaps = (
            #     session.query(OverlapDB)
            #     .join(NodeDB, NodeDB.id == OverlapDB.node_id)
            #     .where(NodeDB.t == time)
            # )

        self._nodes = {node.id: node for node in self._nodes}

        if len(self._nodes) == 0:
            raise ValueError(f"Could not find segmentations at time {time}")

        area = np.asarray([node.area for node in self._nodes.values()])
        self._area_threshold_w.min = area.min()
        self._area_threshold_w.max = area.max()
        self._area_threshold_w.value = np.median(area)

    def _on_threshold_update(self, value: float) -> None:
        segmentation = self._get_segmentation(threshold=value)
        if self._segm_layer_name in self._viewer.layers:
            self._viewer.layers[self._segm_layer_name].data = segmentation
        else:
            layer = self._viewer.add_labels(segmentation, name=self._segm_layer_name)
            layer.mouse_move_callbacks.append(self._on_mouse_move)

    def _get_segmentation(self, threshold: float) -> np.ndarray:
        """
        NOTE:
        when making this interactive it could be interesting to use the overlap data
        to avoid empty regions when visualizing segments
        """
        if self._ndim == 0:
            raise ValueError(
                "Could not find `shape` metadata. It should be saved during `segmentation` on your `workdir`."
            )

        seen_hierarchies = set()

        buffer = np.zeros(self._shape[1:], dtype=np.uint32)  # ignoring time
        for node, hier_id in zip(self._nodes.values(), self._hier_ids):
            if node.area <= threshold or hier_id not in seen_hierarchies:
                # paint segments larger than threshold on empty regions
                node.paint_buffer(buffer, node.id, include_time=False)
                seen_hierarchies.add(hier_id)

        return buffer

    @property
    def _time(self) -> None:
        available_ndim = self._viewer.dims.ndim
        if available_ndim < self._ndim:
            warn(
                "Napari `ndims` smaller than dataset `ndims`. "
                f"Expected {self._ndim}, found {available_ndim}. Using time = 0"
            )
            return 0

        return self._viewer.dims.point[-self._ndim]

    def _on_mouse_move(self, layer: Optional[Labels], event) -> None:
        if not self._link_w.value:
            return
        self._load_neighbors(layer.get_value(event.position, world=True))

    def _load_neighbors(self, index: int) -> None:
        if index is None or index <= 0:
            return

        index = int(index)  # might be numpy array

        LOG.info(f"Loading node index = {index}")

        engine = sqla.create_engine(self.config.database_path, echo=True)
        with Session(engine) as session:
            query = session.query(NodeDB.z, NodeDB.y, NodeDB.x, LinkDB.weight).where(
                LinkDB.target_id == NodeDB.id, LinkDB.source_id == index
            )
            df = pd.read_sql(query.statement, session.bind)

        LOG.info(f"Found {len(df)} neighbors")

        if len(df) == 0:
            return

        node = self._nodes[index]
        ndim = len(node.centroid)
        centroids = df[["z", "y", "x"]].values[:, -ndim:]  # removing z if 2D

        vectors = np.tile(self._nodes[index].centroid, (len(df), 2, 1))
        vectors[:, 1, :] = centroids - vectors[:, 0, :]

        if self._link_layer_name in self._viewer.layers:
            self._viewer.layers.remove(self._link_layer_name)

        self._viewer.add_vectors(
            data=vectors,
            name=self._link_layer_name,
            features={"weights": df["weight"]},
            edge_color="weights",
            opacity=1.0,
        )

        LOG.info(f"vectors:\n{vectors}")

        self._viewer.layers.selection.active = self._viewer.layers[
            self._segm_layer_name
        ]



import click

from ultrack.cli.check_gurobi import check_gurobi_cli
from ultrack.cli.clear_database import clear_database_cli
from ultrack.cli.config import config_cli
from ultrack.cli.data_summary import data_summary_cli
from ultrack.cli.estimate_params import estimate_params_cli
from ultrack.cli.export import export_cli
from ultrack.cli.flow import add_flow_cli
from ultrack.cli.labels_to_edges import labels_to_contours_cli
from ultrack.cli.link import link_cli
from ultrack.cli.match_gt import match_gt_cli
from ultrack.cli.predict import add_probs_cli
from ultrack.cli.segment import segmentation_cli
from ultrack.cli.server import server_cli
from ultrack.cli.solve import solve_cli
from ultrack.cli.view import view_cli


@click.group()
def main():
    pass


main.add_command(add_flow_cli)
main.add_command(check_gurobi_cli)
main.add_command(clear_database_cli)
main.add_command(config_cli)
main.add_command(data_summary_cli)
main.add_command(estimate_params_cli)
main.add_command(export_cli)
main.add_command(labels_to_contours_cli)
main.add_command(link_cli)
main.add_command(match_gt_cli)
main.add_command(add_probs_cli)
main.add_command(segmentation_cli)
main.add_command(solve_cli)
main.add_command(server_cli)
main.add_command(view_cli)



import xml.dom.minidom as minidom
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Optional, Union

import pandas as pd

from ultrack.config.config import MainConfig
from ultrack.core.export.tracks_layer import to_tracks_layer
from ultrack.utils.constants import NO_PARENT


def _set_filter_elem(elem: ET.Element) -> None:
    elem.set("feature", "QUALITY")
    elem.set("value", "0.0")
    elem.set("isabove", "true")


def tracks_layer_to_trackmate(
    tracks_df: pd.DataFrame,
) -> str:
    """
    Convert a pandas DataFrame representation of Napari track layer to TrackMate XML format.
    `<ImageData/>` need to be set manually in the output XML.

    Parameters
    ----------
    tracks_df : pd.DataFrame
        A DataFrame with columns `track_id, id, parent_id, t, z, y, x`. Cells that belong to the same track
        have the same `track_id`.

    Returns
    -------
    str
        A string representation of the XML in the TrackMate format.

    Examples
    --------
    >>> tracks_df = pd.DataFrame(
    ...     [[1,0,12.0,49.0,49.0,1000001,-1,-1],
    ...     [1,1,12.0,49.0,32.0,2000001,-1,1000001],
    ...     [2,1,12.0,49.0,66.0,2000002,-1,1000001]],
    ...     columns=['track_id','t','z','y','x','id','parent_track_id','parent_id']
    ... )
    >>> print(tracks_df)
       track_id  t     z     y     x       id  parent_track_id  parent_id
    0         1  0  12.0  49.0  49.0  1000001               -1         -1
    1         1  1  12.0  49.0  32.0  2000001               -1    1000001
    2         2  1  12.0  49.0  66.0  2000002               -1    1000001
    >>> tracks_layer_to_trackmate(tracks_df)
    <?xml version="1.0" ?>
    <TrackMate version="7.11.1">
        <Model spatialunits="pixels" timeunits="frames">
            <AllTracks>
                <Track TRACK_ID="1" NUMBER_SPOTS="2" NUMBER_GAPS="0" TRACK_START="0" TRACK_STOP="1" name="Track_1">
                    <Edge SPOT_SOURCE_ID="1000001" SPOT_TARGET_ID="2000001" EDGE_TIME="0.5"/>
                </Track>
                <Track TRACK_ID="2" NUMBER_SPOTS="1" NUMBER_GAPS="0" TRACK_START="1" TRACK_STOP="1" name="Track_2">
                    <Edge SPOT_SOURCE_ID="1000001" SPOT_TARGET_ID="2000002" EDGE_TIME="0.5"/>
                </Track>
            </AllTracks>
            <FilteredTracks>
                <TrackID TRACK_ID="1"/>
                <TrackID TRACK_ID="2"/>
            </FilteredTracks>
            <AllSpots>
                <SpotsInFrame frame="0">
                    <Spot ID="1000001" QUALITY="1.0" VISIBILITY="1" name="1000001" FRAME="0" RADIUS="5.0" POSITION_X="49.0" POSITION_Y="49.0" POSITION_Z="12.0"/>
                </SpotsInFrame>
                <SpotsInFrame frame="1">
                    <Spot ID="2000001" QUALITY="1.0" VISIBILITY="1" name="2000001" FRAME="1" RADIUS="5.0" POSITION_X="32.0" POSITION_Y="49.0" POSITION_Z="12.0"/>
                    <Spot ID="2000002" QUALITY="1.0" VISIBILITY="1" name="2000002" FRAME="1" RADIUS="5.0" POSITION_X="66.0" POSITION_Y="49.0" POSITION_Z="12.0"/>
                </SpotsInFrame>
            </AllSpots>
            <FeatureDeclarations>
                ...
            </FeatureDeclarations>
        </Model>
        <Settings>
            <InitialSpotFilter feature="QUALITY" value="0.0" isabove="true"/>
            <SpotFilterCollection/>
            <TrackFilterCollection/>
            <ImageData filename="None" folder="None" width="0" height="0" depth="0" nslices="1" nframes="2" pixelwidth="1.0" pixelheight="1.0" voxeldepth="1.0" timeinterval="1.0"/>
        </Settings>
    </TrackMate>
    """  # noqa: E501
    tracks_df["id"] = tracks_df["id"].astype(int)
    if not tracks_df["id"].is_unique:
        raise ValueError("The 'id' column must be unique.")
    tracks_df.set_index("id", inplace=True)
    tracks_df["parent_id"] = tracks_df["parent_id"].astype(int)
    tracks_df["track_id"] = tracks_df["track_id"].astype(int)

    # Create XML root and child elements
    root = ET.Element("TrackMate")
    root.set("version", "7.11.1")  # required by TrackMate, not significant

    model_elem = ET.SubElement(root, "Model")
    model_elem.set("spatialunits", "pixels")
    model_elem.set("timeunits", "frames")

    all_tracks_elem = ET.SubElement(model_elem, "AllTracks")
    filtered_tracks_elem = ET.SubElement(model_elem, "FilteredTracks")
    all_spots_elem = ET.SubElement(model_elem, "AllSpots")
    features_elem = ET.SubElement(model_elem, "FeatureDeclarations")

    settings_elem = ET.SubElement(root, "Settings")

    _set_filter_elem(ET.SubElement(settings_elem, "InitialSpotFilter"))
    ET.SubElement(settings_elem, "SpotFilterCollection")
    ET.SubElement(settings_elem, "TrackFilterCollection")

    image_elem = ET.SubElement(settings_elem, "ImageData")
    image_elem.set("filename", "None")
    image_elem.set("folder", "None")
    image_elem.set("width", "0")
    image_elem.set("height", "0")
    image_elem.set("depth", "0")
    image_elem.set("nslices", "1")
    image_elem.set("nframes", str(tracks_df["t"].max() + 1))
    image_elem.set("pixelwidth", "1.0")
    image_elem.set("pixelheight", "1.0")
    image_elem.set("voxeldepth", "1.0")
    image_elem.set("timeinterval", "1.0")

    has_z = "z" in tracks_df.columns

    # Create spot features
    spot_features_elem = ET.SubElement(features_elem, "SpotFeatures")
    spot_features = [
        ("QUALITY", "Quality", "Quality", "QUALITY", "false"),
        ("POSITION_X", "X", "X", "POSITION", "false"),
        ("POSITION_Y", "Y", "Y", "POSITION", "false"),
        ("POSITION_Z", "Z", "Z", "POSITION", "false"),
        ("POSITION_T", "T", "T", "TIME", "false"),
        ("FRAME", "Frame", "Frame", "NONE", "true"),
        ("RADIUS", "Radius", "R", "LENGTH", "false"),
        ("VISIBILITY", "Visibility", "Visibility", "NONE", "true"),
        ("MANUAL_COLOR", "Manual spot color", "Spot color", "NONE", "true"),
        ("MEAN_INTENSITY", "Mean intensity", "Mean", "INTENSITY", "false"),
        ("MEDIAN_INTENSITY", "Median intensity", "Median", "INTENSITY", "false"),
        ("MIN_INTENSITY", "Minimal intensity", "Min", "INTENSITY", "false"),
        ("MAX_INTENSITY", "Maximal intensity", "Max", "INTENSITY", "false"),
        ("TOTAL_INTENSITY", "Total intensity", "Total int.", "INTENSITY", "false"),
        ("STANDARD_DEVIATION", "Standard deviation", "Stdev.", "INTENSITY", "false"),
        ("ESTIMATED_DIAMETER", "Estimated diameter", "Diam.", "LENGTH", "false"),
        ("CONTRAST", "Contrast", "Constrast", "NONE", "false"),
        ("SNR", "Signal/Noise ratio", "SNR", "NONE", "false"),
    ]
    for feature, name, shortname, dimension, isint in spot_features:
        elem = ET.SubElement(spot_features_elem, "Feature")
        elem.set("feature", feature)
        elem.set("name", name)
        elem.set("shortname", shortname)
        elem.set("dimension", dimension)
        elem.set("isint", isint)

    # Create edge features
    edge_features_elem = ET.SubElement(features_elem, "EdgeFeatures")
    edge_features = [
        ("SPOT_SOURCE_ID", "Source spot ID", "Source ID", "NONE", "true"),
        ("SPOT_TARGET_ID", "Target spot ID", "Target ID", "NONE", "true"),
        # ... add other edge features if needed
    ]
    for feature, name, shortname, dimension, isint in edge_features:
        elem = ET.SubElement(edge_features_elem, "Feature")
        elem.set("feature", feature)
        elem.set("name", name)
        elem.set("shortname", shortname)
        elem.set("dimension", dimension)
        elem.set("isint", isint)

    track_features_elem = ET.SubElement(features_elem, "TrackFeatures")
    track_features = [
        ("NUMBER_SPOTS", "Number of spots in track", "N spots", "NONE", "true"),
        ("NUMBER_GAPS", "Number of gaps", "Gaps", "NONE", "true"),
        ("LONGEST_GAP", "Longest gap", "Longest gap", "NONE", "true"),
        ("NUMBER_SPLITS", "Number of split events", "Splits", "NONE", "true"),
        ("NUMBER_MERGES", "Number of merge events", "Merges", "NONE", "true"),
        ("NUMBER_COMPLEX", "Complex points", "Complex", "NONE", "true"),
        ("TRACK_DURATION", "Duration of track", "Duration", "TIME", "false"),
        ("TRACK_START", "Track start", "T start", "TIME", "false"),
        ("TRACK_STOP", "Track stop", "T stop", "TIME", "false"),
        ("TRACK_DISPLACEMENT", "Track displacement", "Displacement", "LENGTH", "false"),
        ("TRACK_INDEX", "Track index", "Index", "NONE", "true"),
        ("TRACK_ID", "Track ID", "ID", "NONE", "true"),
    ]

    for feature, name, shortname, dimension, isint in track_features:
        elem = ET.SubElement(track_features_elem, "Feature")
        elem.set("feature", feature)
        elem.set("name", name)
        elem.set("shortname", shortname)
        elem.set("dimension", dimension)
        elem.set("isint", isint)

    # Create spots
    for frame, group in tracks_df.groupby("t"):
        frame_elem = ET.SubElement(all_spots_elem, "SpotsInFrame")
        frame_elem.set("frame", str(frame))
        for spot_id, entry in group.iterrows():
            spot_elem = ET.SubElement(frame_elem, "Spot")
            spot_elem.set("ID", str(spot_id))
            spot_elem.set("QUALITY", "1.0")
            spot_elem.set("VISIBILITY", "1")
            spot_elem.set("name", str(spot_id))
            spot_elem.set("FRAME", str(int(entry["t"])))
            spot_elem.set("RADIUS", "5.0")
            spot_elem.set("POSITION_T", str(entry["t"]))
            spot_elem.set("POSITION_X", str(entry["x"]))
            spot_elem.set("POSITION_Y", str(entry["y"]))
            # features with default values
            spot_elem.set("MANUAL_COLOR", "1"),
            spot_elem.set("MEAN_INTENSITY", "1.0"),
            spot_elem.set("MEDIAN_INTENSITY", "1.0"),
            spot_elem.set("MIN_INTENSITY", "1.0"),
            spot_elem.set("MAX_INTENSITY", "1.0"),
            spot_elem.set("TOTAL_INTENSITY", "1.0"),
            spot_elem.set("STANDARD_DEVIATION", "1.0"),
            spot_elem.set("ESTIMATED_DIAMETER", "10.0"),
            spot_elem.set("CONTRAST", "0.0"),
            spot_elem.set("SNR", "0.0"),
            if has_z:
                spot_elem.set("POSITION_Z", str(entry["z"]))
            else:
                spot_elem.set("POSITION_Z", "0.0")

    # Create tracks using lineage
    for track_id, group in tracks_df.groupby("track_id"):
        track_elem = ET.SubElement(all_tracks_elem, "Track")
        track_elem.set("TRACK_ID", str(track_id))
        track_elem.set("NUMBER_SPOTS", str(len(group)))
        track_elem.set("NUMBER_GAPS", "0")
        track_elem.set("TRACK_START", str(group["t"].min()))
        track_elem.set("TRACK_STOP", str(group["t"].max()))
        track_elem.set("name", f"Track_{track_id}")

        ET.SubElement(filtered_tracks_elem, "TrackID").set("TRACK_ID", str(track_id))

        for spot_id, entry in group.iterrows():  # spot_id is the DataFrame row index
            parent_id = int(entry["parent_id"])
            if parent_id == NO_PARENT:
                continue
            edge_elem = ET.SubElement(track_elem, "Edge")
            edge_elem.set("SPOT_SOURCE_ID", str(parent_id))
            edge_elem.set("SPOT_TARGET_ID", str(spot_id))
            edge_elem.set("EDGE_TIME", str(entry["t"] - 0.5))

    # Convert to XML string
    xml_str = ET.tostring(root, encoding="unicode")
    xml_str = minidom.parseString(xml_str).toprettyxml()

    return xml_str


def to_trackmate(
    config: MainConfig,
    output_path: Optional[Union[Path, str]] = None,
    overwrite: bool = False,
) -> str:
    """
    Exports tracking results to TrackMate XML format.

    Parameters
    ----------
    config : MainConfig
        ULTrack configuration parameters.
    output_path : Optional[Path], optional
        Output file path, by default None
    overwrite : bool, optional
        Whether to overwrite the output file if it already exists, by default False

    Returns
    -------
    str
        A string representation of the XML in the TrackMate format.
    """
    tracks_df, _ = to_tracks_layer(config)
    xml_str = tracks_layer_to_trackmate(tracks_df)

    # Save to file if output_path is provided
    if output_path is not None:
        if isinstance(output_path, str):
            output_path = Path(output_path)

        if output_path.exists() and not overwrite:
            raise FileExistsError(
                f"File {output_path} already exists. Set overwrite=True to overwrite."
            )

        with output_path.open("w") as f:
            f.write(xml_str)

    return xml_str



import shutil
from pathlib import Path
from typing import Callable, List, Optional, Tuple, Union
from uuid import uuid4

import numpy as np
import scipy.ndimage as ndi
import torch as th
from numpy.typing import ArrayLike
from skimage.feature import peak_local_max
from skimage.segmentation import find_boundaries

from ultrack.utils.cuda import torch_default_device


class MicroSAM:
    """
    Using a SAM model, generates masks for the entire image.
    Generates a grid of point prompts over the image, then filters
    low quality and duplicate masks. The default settings are chosen
    for SAM with a ViT-H backbone.

    Parameters
    ----------
    device : Union[str, th.device, None] = None,
        The device to run the model on. If None, defaults to the default.
    points_per_side : int or None, optional
        The number of points to be sampled along one side of the image.
        The total number of points is points_per_side**2. If None, 'point_grids'
        must provide explicit point sampling.
    points_per_batch : int, optional
        Sets the number of points run simultaneously by the model.
        Higher numbers may be faster but use more GPU memory.
    pred_iou_thresh : float, optional
        A filtering threshold in [0,1], using the model's predicted mask quality.
    stability_score_thresh : float, optional
        A filtering threshold in [0,1], using the stability of the mask under
        changes to the cutoff used to binarize the model's mask predictions.
    stability_score_offset : float, optional
        The amount to shift the cutoff when calculating the stability score.
    box_nms_thresh : float, optional
        The box IoU cutoff used by non-maximal suppression to filter duplicate masks.
    crop_n_layers : int, optional
        If >0, mask prediction will be run again on crops of the image. Sets the
        number of layers to run, where each layer has 2**i_layer number of image crops.
    crop_nms_thresh : float, optional
        The box IoU cutoff used by non-maximal suppression to filter duplicate
        masks between different crops.
    crop_overlap_ratio : float, optional
        Sets the degree to which crops overlap. In the first crop layer, crops will
        overlap by this fraction of the image length. Later layers with more crops
        scale down this overlap.
    crop_n_points_downscale_factor : int, optional
        The number of points-per-side sampled in layer n is scaled down by
        crop_n_points_downscale_factor**n.
    point_grids : list[np.ndarray] or None, optional
        A list over explicit grids of points used for sampling, normalized to [0,1].
        The nth grid in the list is used in the nth crop layer. Exclusive with points_per_side.
    min_mask_region_area : int, optional
        If >0, postprocessing will be applied to remove disconnected regions and holes
        in masks with an area smaller than min_mask_region_area. Requires opencv.
    max_mask_region_prop : float, optional
        Regions larger than max_mask_region_prop * image area will be removed.
    mutex_watershed : bool, optional
        If True, uses a mutex watershed algorithm to generate masks.
    tile_shape : Tuple[int, int], optional
        The shape of the tiles used to precompute image embeddings.
    halo_shape : Tuple[int, int], optional
        The shape of the halo (overlap) around each tile used to precompute image embeddings.
    """

    def __init__(
        self,
        model_type: str = "vit_h_lm",
        device: Union[str, th.device, None] = None,
        points_per_side: Optional[int] = 64,
        points_per_batch: int = 128,
        pred_iou_thresh: float = 0.88,
        stability_score_thresh: float = 0.5,
        stability_score_offset: float = 0.75,
        box_nms_thresh: float = 0.5,
        crop_n_layers: int = 0,
        crop_nms_thresh: float = 0.7,
        crop_overlap_ratio: float = 512 / 1500,
        crop_n_points_downscale_factor: int = 1,
        point_grids: Optional[List[np.ndarray]] = None,
        min_mask_region_area: int = 0,
        max_mask_region_prop: float = 0.1,
        mutex_watershed: bool = False,
        tile_shape: Tuple[int, int] = (512, 512),
        halo_shape: Tuple[int, int] = (128, 128),
    ) -> None:
        from micro_sam import instance_segmentation, util

        if device is None:
            device = torch_default_device()

        if isinstance(device, str):
            device = th.device(device)

        self._predictor = util.get_sam_model(device=device, model_type=model_type)

        self._mutex_watershed = mutex_watershed
        if mutex_watershed:
            self._amg = instance_segmentation.TiledEmbeddingMaskGenerator(
                self._predictor,
                n_threads=1,
                stability_score_offset=stability_score_offset,
            )
        else:
            self._amg = instance_segmentation.TiledAutomaticMaskGenerator(
                self._predictor,
                points_per_side=points_per_side,
                points_per_batch=points_per_batch,
                point_grids=point_grids,
                stability_score_offset=stability_score_offset,
            )

        self._pred_iou_thresh = pred_iou_thresh
        self._stability_score_thresh = stability_score_thresh
        self._box_nms_thresh = box_nms_thresh
        self._crop_n_layers = crop_n_layers
        self._crop_nms_thresh = crop_nms_thresh
        self._crop_overlap_ratio = crop_overlap_ratio
        self._crop_n_points_downscale_factor = crop_n_points_downscale_factor
        self._min_mask_region_area = min_mask_region_area
        self._max_mask_region_prop = max_mask_region_prop

        self._embedding_cache_dir = Path("/tmp/micro-sam")

        self._tile_shape = tile_shape
        self._halo_shape = halo_shape

    def __call__(self, image: ArrayLike) -> np.ndarray:
        """
        Estimate contour of objects of an image.

        Background regions not assigned to any label are set to -1.

        Parameters
        ----------
        image : ArrayLike
            The input image to be processed.

        Returns
        -------
        np.ndarray
            The processed image with contours derived from the identified masks.
        """
        from micro_sam.util import precompute_image_embeddings
        from segment_anything.utils.amg import area_from_rle, rle_to_mask

        image = np.asarray(image)

        if image.ndim != 2:
            raise ValueError(f"Image must be 2D, got {image.ndim}D.")

        embedding_path = self._embedding_cache_dir / f"{uuid4()}.zarr"

        if embedding_path.exists():
            shutil.rmtree(embedding_path)

        embeddings = precompute_image_embeddings(
            self._predictor,
            image,
            save_path=str(embedding_path),
            tile_shape=self._tile_shape,
            halo=self._halo_shape,
        )

        self._amg.initialize(image, embeddings, verbose=True)

        if self._mutex_watershed:
            preds = self._amg.generate(
                pred_iou_thresh=self._pred_iou_thresh,
                stability_score_thresh=self._stability_score_thresh,
                box_nms_thresh=self._box_nms_thresh,
                min_mask_region_area=self._min_mask_region_area,
            )
            contour = find_boundaries(preds, mode="inner")
            detection = preds > 0

        else:
            preds = self._amg.generate(
                pred_iou_thresh=self._pred_iou_thresh,
                stability_score_thresh=self._stability_score_thresh,
                box_nms_thresh=self._box_nms_thresh,
                crop_nms_thresh=self._crop_nms_thresh,
                min_mask_region_area=self._min_mask_region_area,
                output_mode="uncompressed_rle",
            )

            contour = th.zeros(image.shape, dtype=th.float32)
            detection = th.zeros(image.shape, dtype=bool)
            max_area = np.prod(image.shape, dtype=float) * self._max_mask_region_prop
            for pred in preds:
                if area_from_rle(pred["segmentation"]) > max_area:
                    continue
                mask = rle_to_mask(pred["segmentation"])
                contour += find_boundaries(mask, mode="inner")
                detection[mask.astype(bool)] = True

            contour /= contour.max()
            contour = contour.numpy()

        contour[~detection] = -1

        shutil.rmtree(embedding_path)

        return contour


def set_peak_maxima_prompt(
    sam: MicroSAM,
    sigma: float,
    min_distance: int,
    threshold_rel: float,
    **kwargs,
) -> Callable:
    """
    Configure a function to locate peak maxima in a given image using the MicroSAM framework.
    The function first applies a Gaussian blur to the image, then finds local maxima,
    and finally processes the image using the MicroSAM framework with the identified maxima.

    Parameters
    ----------
    sam : MicroSAM
        A MicroSAM instance used for image processing.
    sigma : float
        Standard deviation for Gaussian filter applied to the image.
    min_distance : int
        Minimum number of pixels separating peaks. Peaks closer than this are considered a single peak.
    threshold_rel : float
        Minimum intensity difference between the peak and the pixels surrounding the peak for it to be
        considered a true peak. It is a relative threshold with respect to the maximum intensity in the image.
    **kwargs :
        Additional keyword arguments to pass to the `peak_local_max` function.

    Returns
    -------
    Callable
        A function that, when given an image, will return an image processed by
        MicroSAM with contours derived from the identified peak maxima.

    Examples
    --------
    >>> sam_instance = ...  # Initialize or provide a MicroSAM instance.
    >>> peak_maxima_function = set_peak_maxima_prompt(
    >>>     sam_instance, sigma=1, min_distance=10, threshold_rel=0.1,
    >>> )
    >>> processed_image = peak_maxima_function(input_image)
    """

    def _peak_maxima_micro_sam(image: ArrayLike) -> np.ndarray:
        """
        Inner function that processes the given image, finds peak maxima,
        and applies the MicroSAM processing with the identified maxima.

        Parameters
        ----------
        image : ArrayLike
            The input image to be processed.

        Returns
        -------
        np.ndarray
            The processed image with contours derived from the identified peak maxima.
        """
        image = np.asarray(image)
        # find maxima coordinates
        blurred_image = ndi.gaussian_filter(image, sigma=sigma)
        coords = peak_local_max(
            blurred_image,
            min_distance=min_distance,
            threshold_rel=threshold_rel,
            **kwargs,
        )
        coords = coords / image.shape

        prev_grid = sam._amg.point_grids
        # set point grid to maxima coordinates
        sam._amg.point_grids = [coords]
        # run sam
        contour = sam(image)
        # reset point grid
        sam._amg.point_grids = prev_grid

        return contour

    return _peak_maxima_micro_sam



import logging

import networkx as nx
import pandas as pd

from ultrack.config.config import MainConfig
from ultrack.core.export.tracks_layer import to_tracks_layer
from ultrack.tracks.graph import _create_tracks_forest
from ultrack.utils.constants import NO_PARENT

LOG = logging.getLogger(__name__)


def tracks_layer_to_networkx(
    tracks_df: pd.DataFrame,
    children_to_parent: bool = False,
) -> nx.DiGraph:
    """
    Convert napari tracks layer tracks dataframe to networkx directed graph.
    By default, the edges are the parent to child relationships.

    Parameters
    ----------
    config : MainConfig
        Configuration parameters.
    children_to_parent : bool
        If set, edges encode child to parent relationships.

    Returns
    -------
    nx.DiGraph
        Networkx graph.
    """
    graph = nx.DiGraph()

    if "id" not in tracks_df.columns:
        LOG.warning(
            "`id` not found in tracks dataframe. Assuming `id` is dataframe index."
        )
    else:
        tracks_df = tracks_df.set_index("id")

    for index, row in tracks_df.to_dict(orient="index").items():
        graph.add_node(index, **row)

    if "parent_id" not in tracks_df.columns:
        LOG.warning(
            "Parent id not found in tracks dataframe. Divisions won't be exported."
        )
        for _, group in tracks_df.groupby("track_id"):
            for i in range(len(group) - 1):
                if children_to_parent:
                    graph.add_edge(group.index[i + 1], group.index[i])
                else:
                    graph.add_edge(group.index[i], group.index[i + 1])

    else:
        dict_graph = _create_tracks_forest(
            tracks_df.index.values,
            tracks_df["parent_id"].to_numpy(int),
        )
        dict_graph.pop(NO_PARENT)

        for parent, children in dict_graph.items():
            for child in children:
                if children_to_parent:
                    graph.add_edge(child, parent)
                else:
                    graph.add_edge(parent, child)

    return graph


def to_networkx(
    config: MainConfig,
    children_to_parent: bool = False,
) -> nx.DiGraph:
    """
    Convert solution from database to networkx directed graph.
    By default, the edges are the parent to child relationships.

    Parameters
    ----------
    config : MainConfig
        Configuration parameters.
    children_to_parent : bool
        If set, edges encode child to parent relationships.

    Returns
    -------
    nx.DiGraph
        Networkx graph.
    """
    df, _ = to_tracks_layer(config)
    return tracks_layer_to_networkx(df, children_to_parent)



import logging

import sqlalchemy as sqla
from sqlalchemy.orm import Session

from ultrack.core.database import LinkDB
from ultrack.core.solve.sqltracking import SQLTracking

LOG = logging.getLogger(__name__)


def clear_linking_data(database_path: str) -> None:
    """Clears linking data."""
    SQLTracking.clear_solution_from_database(database_path)

    LOG.info("Clearing links database.")
    engine = sqla.create_engine(database_path)
    with Session(engine) as session:
        session.query(LinkDB).delete()
        session.commit()



import numpy as np
import pandas as pd
import pytest

from ultrack.tracks import (
    filter_short_sibling_tracks,
    get_paths_to_roots,
    get_subgraph,
    split_trees,
)
from ultrack.utils.constants import NO_PARENT


@pytest.fixture
# Sample data for testing
def tracks_df() -> pd.DataFrame:
    data = {
        "track_id": [1, 2, 3, 4, 5, 6, 7, 8],
        "parent_track_id": [NO_PARENT, 1, 1, 2, 2, NO_PARENT, 6, 6],
    }
    return pd.DataFrame(data)


def test_get_subgraph_single_track_id(tracks_df: pd.DataFrame) -> None:
    track_ids = [3]
    lineage_ids = [1, 2, 3, 4, 5]
    result_df = get_subgraph(tracks_df, track_ids)
    expected_df = tracks_df[tracks_df["track_id"].isin(lineage_ids)]
    pd.testing.assert_frame_equal(result_df, expected_df)


def test_get_subgraph_multiple_track_ids(tracks_df: pd.DataFrame) -> None:
    track_ids = [2, 3, 5]
    lineage_ids = [1, 2, 3, 4, 5]
    result_df = get_subgraph(tracks_df, track_ids)
    expected_df = tracks_df[tracks_df["track_id"].isin(lineage_ids)]
    pd.testing.assert_frame_equal(result_df, expected_df)


def test_get_subgraph_from_different_trees(tracks_df: pd.DataFrame) -> None:
    track_ids = [3, 7]
    result_df = get_subgraph(tracks_df, track_ids)
    pd.testing.assert_frame_equal(result_df, tracks_df)


def test_get_subgraph_empty_subgraph(tracks_df: pd.DataFrame) -> None:
    track_ids = [9, 10, 11]
    result_df = get_subgraph(tracks_df, track_ids)
    assert result_df.empty


def test_get_paths_to_roots() -> None:
    tracks_df = pd.DataFrame(
        {
            "t": [1, 2, 3, 3, 4, 4, 5, 1, 2, 2],
            "track_id": [1, 1, 2, 3, 4, 5, 5, 6, 7, 8],
            "parent_track_id": [NO_PARENT, NO_PARENT, 1, 1, 2, 2, 2, NO_PARENT, 6, 6],
        },
        index=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    )
    path_df = get_paths_to_roots(tracks_df, track_index=5)

    single_path_df = pd.DataFrame(
        {
            "t": [1, 2, 3, 4, 5],
            "track_id": [5] * 5,
            "parent_track_id": [NO_PARENT] * 5,
        },
        index=[1, 2, 3, 6, 7],
    )

    pd.testing.assert_frame_equal(
        path_df,
        single_path_df,
    )

    path_df = get_paths_to_roots(tracks_df, node_index=6)

    single_path_df = pd.DataFrame(
        {
            "t": [1, 2, 3, 4],
            "track_id": [5] * 4,
            "parent_track_id": [NO_PARENT] * 4,
        },
        index=[1, 2, 3, 6],
    )

    pd.testing.assert_frame_equal(
        path_df,
        single_path_df,
    )

    path_df = get_paths_to_roots(tracks_df)
    all_paths_df = pd.DataFrame(
        {
            "t": [1, 2, 3, 1, 2, 3, 4, 1, 2, 3, 4, 5, 1, 2, 1, 2],
            "track_id": [3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 7, 7, 8, 8],
        },
        index=[1, 2, 4, 1, 2, 3, 5, 1, 2, 3, 6, 7, 8, 9, 8, 10],
    )
    all_paths_df["parent_track_id"] = NO_PARENT

    pd.testing.assert_frame_equal(
        path_df,
        all_paths_df,
    )


def test_filter_short_sibling_tracks() -> None:
    """
    Input:
              4 -
             /
        / 3 -      6 ----
        |    \\    /
        |     5 -       8 -
        |         \\    /
    1 --|          7 -
        |              \
        |               9 -
        |
        \\ 2 -


    Output:
            6 ----
           /
    1 ----       8 -
           \\    /
            7 -
                \
                 9 -
    """
    segm = np.zeros((8, 6), dtype=np.uint8)

    tracks_df = pd.DataFrame(
        {
            "t": [0, 1, 2, 2, 3, 3, 4, 5, 6, 7, 4, 5, 5],
            "x": [2, 2, 0, 4, 5, 3, 4, 4, 4, 4, 2, 3, 1],
            "track_id": [1, 1, 2, 3, 4, 5, 6, 6, 6, 6, 7, 8, 9],
            "parent_track_id": [NO_PARENT, NO_PARENT, 1, 1, 3, 3, 5, 5, 5, 5, 5, 7, 7],
        }
    )
    segm[tracks_df["t"].astype(int), tracks_df["x"].astype(int)] = tracks_df[
        "track_id"
    ].values

    filtered_df, filtered_segm = filter_short_sibling_tracks(
        tracks_df, min_length=3, segments=segm
    )
    expected_df = pd.DataFrame(
        {
            "t": [0, 1, 2, 3, 4, 5, 6, 7, 4, 5, 5],
            "x": [2, 2, 4, 3, 4, 4, 4, 4, 2, 3, 1],
            "track_id": [1, 1, 1, 1, 6, 6, 6, 6, 7, 8, 9],
            "parent_track_id": [
                NO_PARENT,
                NO_PARENT,
                NO_PARENT,
                NO_PARENT,
                1,
                1,
                1,
                1,
                1,
                7,
                7,
            ],
        },
        index=[0, 1, 3, 5, 6, 7, 8, 9, 10, 11, 12],
    )
    expected_segm = np.zeros((8, 6), dtype=np.uint8)
    expected_segm[expected_df["t"].astype(int), expected_df["x"].astype(int)] = (
        expected_df["track_id"].values
    )

    pd.testing.assert_frame_equal(
        filtered_df,
        expected_df,
    )
    np.testing.assert_array_equal(
        filtered_segm,
        expected_segm,
    )


def test_split_trees() -> None:
    tracks_df = pd.DataFrame(
        {
            "track_id": [1, 2, 3, 4, 4],
            "parent_track_id": [NO_PARENT, 1, 1, NO_PARENT, NO_PARENT],
        }
    )

    forest = split_trees(tracks_df)
    expected_forest = [
        pd.DataFrame(
            {
                "track_id": [1, 2, 3],
                "parent_track_id": [NO_PARENT, 1, 1],
            },
            index=[0, 1, 2],
        ),
        pd.DataFrame(
            {
                "track_id": [4, 4],
                "parent_track_id": [NO_PARENT, NO_PARENT],
            },
            index=[3, 4],
        ),
    ]

    for tree, expected_tree in zip(forest, expected_forest):
        pd.testing.assert_frame_equal(
            tree,
            expected_tree,
        )



from typing import List

import napari
from sqlalchemy import create_engine, func
from sqlalchemy.orm import Session

from ultrack.core.database import (
    NO_PARENT,
    NodeDB,
    VarAnnotation,
    get_node_values,
    set_node_values,
)
from ultrack.core.segmentation.node import Node
from ultrack.widgets._generic_annotation_widget import GenericAnnotationWidget


class DivisionAnnotationWidget(GenericAnnotationWidget):
    def __init__(self, viewer: napari.Viewer) -> None:
        # before init due to config initialization
        super().__init__(
            viewer, "Division Annotation", VarAnnotation, " ~ division annot."
        )

    def _query_samples(self) -> List[Node]:
        engine = create_engine(self.config.database_path)
        with Session(engine) as session:
            parent_ids = (
                session.query(NodeDB.parent_id)
                .where(
                    NodeDB.parent_id != NO_PARENT,
                    NodeDB.parent_id.not_in([node.id for node in self._nodes]),
                )
                .group_by(NodeDB.parent_id)
                .having(func.count(NodeDB.id) > 1)
                .order_by(func.random())
                .limit(self._sample_size)
            )
            nodes = session.query(NodeDB.pickle).where(NodeDB.id.in_(parent_ids))

        return [node for node, in nodes]

    def get_annotation(self, index: int) -> VarAnnotation:
        return get_node_values(self.config, index, NodeDB.division)

    def set_annotation(self, index: int, annot: VarAnnotation) -> None:
        set_node_values(self.config, index, division=annot)



from ultrack.core.export.ctc import to_ctc
from ultrack.core.export.exporter import export_tracks_by_extension
from ultrack.core.export.geff import to_geff
from ultrack.core.export.networkx import to_networkx, tracks_layer_to_networkx
from ultrack.core.export.trackmate import to_trackmate, tracks_layer_to_trackmate
from ultrack.core.export.tracks_layer import to_tracks_layer
from ultrack.core.export.zarr import tracks_to_zarr



from pathlib import Path
from typing import Callable

import napari
import numpy as np
import pytest

from ultrack import to_tracks_layer
from ultrack.config import MainConfig
from ultrack.widgets import TrackInspectionWidget

pytest.skip(reason="DEPRECATED", allow_module_level=True)


@pytest.mark.parametrize(
    "config_content,timelapse_mock_data",
    [
        (
            {
                "segmentation.n_workers": 4,
                "linking.n_workers": 4,
                "tracking.appear_weight": -0.25,
                "tracking.disappear_weight": -0.5,
                "tracking.division_weight": -0.25,
            },
            {"length": 4},
        )
    ],
    indirect=True,
)
def test_division_annotation_widget(
    make_napari_viewer: Callable[[], napari.Viewer],
    tracked_database_mock_data: MainConfig,
    tmp_path: Path,
    request,
) -> None:
    # NOTE: Use "--show-napari-viewer" to show viewer, useful when debugging

    config = tracked_database_mock_data
    config.data_config.working_dir = tmp_path

    viewer = make_napari_viewer()
    widget = TrackInspectionWidget(viewer)
    viewer.window.add_dock_widget(widget)

    assert not widget._next_btn.enabled
    assert not widget._prev_btn.enabled

    tracks, graph = to_tracks_layer(config.data_config)
    tracks = tracks.to_numpy()
    tracks[:, 2:] += np.random.uniform(3, size=tracks[:, 2:].shape)

    tracks_layer = viewer.add_tracks(tracks, graph=graph)

    if request.config.getoption("--show-napari-viewer"):
        napari.run()
        return

    assert widget._tracks_layer_w.value == tracks_layer
    assert len(widget._sorted_tracks) > 0

    next_count = 0
    while widget._next_btn.enabled:
        widget._next_btn.clicked()
        next_count += 1
        assert widget._current_track_layer is not None

    prev_count = 0
    while widget._prev_btn.enabled:
        widget._prev_btn.clicked()
        prev_count += 1
        assert widget._current_track_layer is not None

    assert next_count > 0
    assert prev_count == next_count



from typing import Tuple

import zarr

from ultrack import solve
from ultrack.config import MainConfig
from ultrack.core.match_gt import match_to_ground_truth


def test_match_to_ground_truth(
    linked_database_mock_data: MainConfig,
    timelapse_mock_data: Tuple[zarr.Array, zarr.Array, zarr.Array],
) -> None:

    config = linked_database_mock_data

    _, _, gt = timelapse_mock_data

    df_gt, opt_cfg = match_to_ground_truth(
        config,
        gt,
        track_id_graph={},
        is_segmentation=True,
        optimize_config=True,
    )

    for c in ["gt_track_id", "gt_parent_track_id"]:
        assert c in df_gt.columns

    assert config.linking_config.max_distance != opt_cfg.linking_config.max_distance
    assert config.segmentation_config.min_area != opt_cfg.segmentation_config.min_area
    assert config.segmentation_config.max_area != opt_cfg.segmentation_config.max_area
    assert (
        config.segmentation_config.min_frontier
        != opt_cfg.segmentation_config.min_frontier
    )

    solve(config, use_ground_truth_match=True)



import click
import napari

from ultrack.cli.utils import config_option
from ultrack.config import MainConfig
from ultrack.widgets.hierarchy_viz_widget import HierarchyVizWidget


@click.group("view")
def view_cli() -> None:
    """View data in napari."""


@view_cli.command("hierarchy")
@config_option()
def view_hierarchy_cli(config: MainConfig) -> None:
    """Opens napari viewer with hierarchy visualization widget."""
    viewer = napari.Viewer()
    widget = HierarchyVizWidget(viewer, config)
    viewer.window.add_dock_widget(widget)
    napari.run()



from pathlib import Path
from typing import Union

from pydantic_settings import BaseSettings

from ultrack.config import DataConfig


class Settings(BaseSettings):
    """Settings for the API.

    Attributes
    ----------
    api_results_path : Union[str, Path, None]
        Path to the API results folder. It stores any file/folder of the results from
        intermediate processing steps and final results. Defaults to "/tmp".
    ultrack_data_config : DataConfig
        Ultrack data configuration. It is used to store the configure the database
        connection and the working directory. Defaults to DataConfig().
    """

    api_results_path: Union[str, Path, None] = Path("/tmp")
    ultrack_data_config: DataConfig = DataConfig()


settings = Settings()



import logging
from typing import Optional

import numpy as np
from numpy.typing import ArrayLike

from ultrack.imgproc.utils import _channel_iterator, _parse_voxel_size
from ultrack.utils.cuda import import_module, to_cpu, xp

LOG = logging.getLogger(__name__)


def normalize(
    image: ArrayLike,
    gamma: float,
    lower_q: float = 0.001,
    upper_q: float = 0.9999,
) -> ArrayLike:
    """
    Normalize image to between [0, 1] and applies a gamma transform (x ^ gamma).

    Parameters
    ----------
    image : ArrayLike
        Images as an T,Y,X,C array.
    gamma : float
        Expoent of gamma transform.
    lower_q : float, optional
        Lower quantile for normalization.
    upper_q : float, optional
        Upper quantile for normalization.

    Returns
    -------
    ArrayLike
        Normalized array.
    """
    frame = image
    frame = frame - np.quantile(frame, lower_q)
    frame = frame / np.quantile(frame, upper_q)
    frame = np.clip(frame, 0, 1)

    if gamma != 1.0:
        frame = np.power(frame, gamma)

    return frame


def robust_invert(
    image: ArrayLike,
    voxel_size: Optional[ArrayLike] = None,
    sigma: float = 1.0,
    lower_quantile: Optional[float] = None,
    upper_quantile: Optional[float] = 0.9999,
    channel_axis: Optional[int] = None,
) -> np.ndarray:
    """
    Inverts an image robustly by first smoothing it with a gaussian filter
    and then normalizing it to [0, 1].

    Parameters
    ----------
    image : ArrayLike
        Input image.
    voxel_size : ArrayLike
        Array of voxel size (z, y, x).
    sigma : float, optional
        Sigma used to smooth the image, by default 1.0.
    lower_quantile : Optional[float], optional
        Lower quantile used to clip the intensities, minimum used when None.
    upper_quantile : Optional[float], optional
        Upper quantile used to clip the intensities, maximum used when None.
    channel_axis : Optional[int], optional
        When provided it will invert each channel separately and merge them.

    Returns
    -------
    ArrayLike
        Inverted and normalized image.
    """
    LOG.info(f"Channel axis {channel_axis}")

    ndi = import_module("scipy", "ndimage")

    shape = list(image.shape)
    if channel_axis is not None:
        shape.pop(channel_axis)
        LOG.info(f"Shape={shape} after removing channel axis={channel_axis}")

    voxel_size = _parse_voxel_size(voxel_size, image.ndim, channel_axis)
    sigmas = sigma / voxel_size
    LOG.info(f"Inverting with voxel size {voxel_size} and sigma {sigma}")
    LOG.info(f"Sigmas after scaling {sigmas}")

    iterator = _channel_iterator(image, channel_axis)
    output = np.zeros(shape, dtype=np.float32)
    norm_factor = 1 / len(iterator)

    for ch in iterator:
        if ch is None:
            _image = image
        else:
            _image = np.take(image, indices=ch, axis=channel_axis)

        _image = xp.asarray(_image)
        _image = ndi.gaussian_filter(_image, sigma=sigmas)

        flat_small_img = ndi.zoom(_image, (0.25,) * _image.ndim, order=1).ravel()

        if lower_quantile is not None:
            im_min = np.quantile(flat_small_img, lower_quantile)
        else:
            im_min = flat_small_img.min()

        im_min = im_min.astype(_image.dtype)
        # safe sub, it could be unsigned
        np.clip(_image, im_min, None, out=_image)
        _image -= im_min
        LOG.info(f"Minimum {im_min}")

        if upper_quantile is not None:
            im_max = np.quantile(flat_small_img, upper_quantile)
        else:
            im_max = flat_small_img.max()
        del flat_small_img

        im_max = np.maximum(im_max.astype(np.float32), 1e-8)
        _image = (1 / im_max) * norm_factor * _image
        LOG.info(f"Maximum {im_max}")

        np.subtract(norm_factor, _image, out=_image)  # inverting
        np.clip(_image, 0, norm_factor, out=_image)

        output += to_cpu(_image)

    return output



from ultrack.config.config import LinkingConfig, MainConfig, load_config
from ultrack.config.dataconfig import DataConfig
from ultrack.config.segmentationconfig import SegmentationConfig
from ultrack.config.trackingconfig import TrackingConfig



from typing import List

import numpy as np
import pytest

from ultrack.utils.cuda import to_cpu
from ultrack.utils.data import make_segmentation_mock_data
from ultrack.utils.edge import labels_to_contours


@pytest.fixture
def multiple_labels(
    n_samples: int = 3,
    length: int = 2,
    size: int = 128,
    n_dim: int = 3,
) -> List[np.ndarray]:
    """Creates a list of sequences of labels."""

    rng = np.random.default_rng(42)
    labels = []

    for _ in range(n_samples):
        lb = np.empty((length,) + (size,) * n_dim, dtype=int)
        for t in range(length):
            _, _, lb[t] = make_segmentation_mock_data(size=size, n_dim=n_dim, rng=rng)
        labels.append(lb)

    return labels


def test_labels_to_contours(multiple_labels: List[np.ndarray]) -> None:
    """Tests merge and convertion of multiple labels into foreground and contours."""

    foreground, _ = to_cpu(labels_to_contours(multiple_labels, sigma=1.5))

    shape = multiple_labels[0].shape

    for lb in multiple_labels:
        for t in range(shape[0]):
            mask = lb[t] > 0
            assert np.all(foreground[t][mask] > 0)



from typing import Optional, Tuple, Union

import dask.array as da
import numpy as np
import torch as th
from numpy.typing import ArrayLike

from ultrack.utils.cuda import import_module, to_cpu, torch_default_device


class PlantSeg:
    """
    A class for performing boundary detection using the Plant-Seg model.
    Plant-Seg documentation for more details, https://github.com/hci-unihd/plant-seg

    Parameters
    ----------
    model_name : str
        Name of the pre-trained segmentation model.
    model_update : bool, optional
        Update the model if True. Default is False.
    device : {str, torch.device}, optional
        Device for model execution. If None, the default device is used.
        Default is None.
    patch : tuple[int], optional
        Patch size for model inference. Default is (80, 160, 160).
    stride_ratio : float, optional
        Stride ratio for patch sampling. Default is 0.75.
    batch_size : int, optional
        Batch size for inference. Default is None.
    preprocess_sigma : float, optional
        Sigma value for Gaussian preprocessing filter. Default is None.
    postprocess_sigma : float, optional
        Sigma value for Gaussian postprocessing filter. Default is None.
    scale_factor : tuple[float], optional
        Scaling factors for input images. Default is None.

    See Also
    --------
    :class:`ArrayPredictor`: Class for making predictions using the PlantSeg model.

    Examples
    --------
    >>> seg_model = PlantSeg(model_name='generic_light_sheet_3D_unet', batch_size=4)
    >>> segmentation_result = seg_model(image_data)
    """

    def __init__(
        self,
        model_name: str,
        model_update: bool = False,
        device: Optional[Union[str, th.device]] = None,
        patch: Tuple[int] = (80, 160, 160),
        stride_ratio: float = 0.75,
        batch_size: Optional[int] = None,
        preprocess_sigma: Optional[float] = None,
        postprocess_sigma: Optional[float] = None,
        scale_factor: Optional[Tuple[float]] = None,
    ) -> None:
        """
        Initialized Plant-Seg model.
        """
        from plantseg.predictions.functional.array_predictor import ArrayPredictor
        from plantseg.predictions.functional.utils import (
            get_model_config,
            get_patch_halo,
        )

        if device is None:
            device = torch_default_device()

        model, model_config, model_path = get_model_config(model_name, model_update)
        patch_halo = get_patch_halo(model_name)
        state = th.load(model_path, map_location="cpu")
        model.load_state_dict(state)

        self.model_name = model_name
        self.patch = patch
        self.stride_ratio = stride_ratio
        self.preprocess_sigma = preprocess_sigma
        self.postprocess_sigma = postprocess_sigma
        self.scale_factor = (
            scale_factor if scale_factor is None else np.asarray(scale_factor)
        )
        self.predictor = ArrayPredictor(
            model=model,
            in_channels=model_config["in_channels"],
            out_channels=model_config["out_channels"],
            device=device,
            patch=patch,
            patch_halo=patch_halo,
            single_batch_mode=False,
            headless=False,
            disable_tqdm=True,
        )

        # overwriting plantseg batch size
        if batch_size is not None:
            self.predictor.batch_size = batch_size

    def __call__(
        self,
        image: ArrayLike,
        transpose: Optional[Tuple[int]] = None,
    ) -> np.ndarray:
        """
        Perform boundary detection using the PlantSeg model.

        Parameters
        ----------
        image : ArrayLike
            Input image data as a numpy, cupy, or Dask array.
        transpose : tuple[int], optional
            Axes permutation for the input image.
            Permutation is applied after pre-processing and before inference and reverted after inference.
            Default is None.

        Returns
        -------
        np.ndarray
            Segmentation boundary probability map as a numpy array.
        """
        from plantseg.dataprocessing import fix_input_shape
        from plantseg.predictions.functional.utils import get_array_dataset

        if isinstance(image, da.Array):
            # avoiding building a large compute graph
            image = image.compute()

        image = fix_input_shape(image)
        orig_shape = image.shape

        ndi = import_module("scipy", "ndimage", image)

        if self.preprocess_sigma is not None:
            image = ndi.gaussian_filter(image, sigma=self.preprocess_sigma)

        if self.scale_factor is not None:
            image = ndi.zoom(image, zoom=self.scale_factor, order=1)

        if transpose is None:
            patch = self.patch
        else:
            # transposes axes
            image = np.transpose(image, axes=transpose)
            patch = [self.patch[i] for i in transpose]

        dataset = get_array_dataset(
            raw=to_cpu(image),  # making sure dataset used for inference is on the CPU
            model_name=self.model_name,
            patch=patch,
            stride_ratio=self.stride_ratio,
            global_normalization=True,
        )
        probs = self.predictor(dataset)[0]

        # reverting transpose
        if transpose:
            image = np.transpose(image, axes=np.argsort(transpose))

        if self.scale_factor is not None:
            probs = np.asarray(probs, like=image)
            inv_factor = tuple(o / s for o, s in zip(orig_shape, probs.shape))
            probs = ndi.zoom(probs, zoom=inv_factor, order=1)

        if self.postprocess_sigma is not None:
            probs = np.asarray(probs, like=image)
            probs = ndi.gaussian_filter(probs, sigma=self.postprocess_sigma)

        return to_cpu(probs)



import json
import logging
from pathlib import Path
from typing import Union

import networkx as nx

from ultrack.config import MainConfig
from ultrack.core.export.geff import to_geff
from ultrack.core.export.networkx import to_networkx
from ultrack.core.export.trackmate import to_trackmate
from ultrack.core.export.tracks_layer import to_tracks_layer
from ultrack.core.export.zarr import tracks_to_zarr

LOG = logging.getLogger(__name__)


def export_tracks_by_extension(
    config: MainConfig, filename: Union[str, Path], overwrite: bool = False
) -> None:
    """
    Export tracks to a file given the file extension.

    Supported file extensions are .xml, .csv, .zarr, .parquet, .dot, .json, and .geff
    - `.xml` exports to a TrackMate compatible XML file.
    - `.csv` exports to a CSV file.
    - `.parquet` exports to a Parquet file.
    - `.zarr` exports the tracks to dense segments in a `zarr` array format.
    - `.geff` exports the tracks to a `zarr` format using the geff standard.
    - `.dot` exports to a Graphviz DOT file.
    - `.json` exports to a networkx JSON file.

    Parameters
    ----------
    filename : str or Path
        The name of the file to save the tracks to.
    config : MainConfig
        The configuration object.
    overwrite : bool, optional
        Whether to overwrite the file if it already exists, by default False.

    See Also
    --------
    to_trackmate :
        Export tracks to a TrackMate compatible XML file.
    to_tracks_layer :
        Export tracks to a CSV file.
    tracks_to_zarr :
        Export tracks to a `zarr` array.
    to_networkx :
        Export tracks to a networkx graph.
    to_geff :
        Export tracks to a geff file.
    """
    filename = Path(filename)
    if filename.exists() and not overwrite:
        raise FileExistsError(
            f"File {filename} already exists. Set `overwrite=True` to overwrite the file"
        )

    file_ext = filename.suffix
    if file_ext.lower() == ".xml":
        to_trackmate(config, filename, overwrite=True)
    elif file_ext.lower() == ".csv":
        df, _ = to_tracks_layer(config, include_parents=True)
        df.to_csv(filename, index=False)
    elif file_ext.lower() == ".zarr":
        df, _ = to_tracks_layer(config)
        tracks_to_zarr(config, df, filename, overwrite=True)
    elif file_ext.lower() == ".geff":
        to_geff(config, filename, overwrite=overwrite)
    elif file_ext.lower() == ".parquet":
        df, _ = to_tracks_layer(config)
        df.to_parquet(filename)
    elif file_ext.lower() == ".dot":
        G = to_networkx(config)
        nx.drawing.nx_pydot.write_dot(G, filename)
    elif file_ext.lower() == ".json":
        G = to_networkx(config)
        json_data = nx.node_link_data(G)
        with open(filename, "w") as f:
            json.dump(json_data, f)
    else:
        raise ValueError(
            f"Unknown file extension: {file_ext}. "
            "Supported extensions are .xml, .csv, .zarr, .geff, .parquet, .dot, and .json."
        )



from typing import List

import napari
from sqlalchemy import create_engine, func
from sqlalchemy.orm import Session

from ultrack.core.database import (
    NodeDB,
    NodeSegmAnnotation,
    get_node_values,
    set_node_values,
)
from ultrack.core.segmentation.node import Node
from ultrack.widgets._generic_annotation_widget import GenericAnnotationWidget


class NodeAnnotationWidget(GenericAnnotationWidget):
    def __init__(self, viewer: napari.Viewer) -> None:
        # before init due to config initialization
        super().__init__(
            viewer, "Node Annotation", NodeSegmAnnotation, " ~ node annot."
        )

    def _query_samples(self) -> List[Node]:
        engine = create_engine(self.config.database_path)
        with Session(engine) as session:
            nodes = (
                session.query(NodeDB.pickle)
                .where(NodeDB.id.not_in([node.id for node in self._nodes]))
                .order_by(func.random())
                .limit(self._sample_size)
                .all()
            )
        return [node for node, in nodes]

    def get_annotation(self, index: int) -> NodeSegmAnnotation:
        return get_node_values(self.config, index, NodeDB.segm_annotation)

    def set_annotation(self, index: int, annot: NodeSegmAnnotation) -> None:
        set_node_values(self.config, index, annotation=annot)



import itertools
import logging
import math
from typing import Optional, Tuple

import numpy as np
import pandas as pd
import sqlalchemy as sqla
from sqlalchemy.orm import Session

from ultrack.config.config import MainConfig
from ultrack.core.database import (
    NO_PARENT,
    GTLinkDB,
    LinkDB,
    NodeDB,
    OverlapDB,
    VarAnnotation,
    maximum_time_from_database,
)
from ultrack.core.solve.solver import MIPSolver
from ultrack.core.solve.solver.base_solver import BaseSolver

logging.basicConfig()
logging.getLogger("sqlachemy.engine").setLevel(logging.INFO)

LOG = logging.getLogger(__name__)


class SQLTracking:
    def __init__(
        self,
        config: MainConfig,
    ) -> None:
        """
        Helper class to query data from SQL database and dispatch do solver.

        Parameters
        ----------
        config : MainConfig
            Configuration parameters.
        """
        LOG.info(f"SQLTracking with TrackingConfig:\n{config.tracking_config}")

        self._tracking_config = config.tracking_config
        self._data_config = config.data_config
        self._solver: Optional[MIPSolver] = None

        self._max_t = maximum_time_from_database(self._data_config)
        if self._tracking_config.window_size is None:
            LOG.info(f"Window size not set, configured to {self._max_t + 1}.")
            self._window_size = self._max_t + 1
        else:
            self._window_size = self._tracking_config.window_size

        self.num_batches = math.ceil(self._max_t / self._window_size)

    def construct_model(self, index: int = 0) -> None:
        """
        Constructs ILP model for a given batch index.
        Adding nodes, edges and slack variables, and biological, boundary and overlap constraints.

        Parameters
        ----------
        index : int
            Batch index, by default 0, which works for single batch tracking.
        """
        if index < 0 or index >= self.num_batches:
            raise ValueError(
                f"Invalid index {index}, expected between [0, {self.num_batches})."
            )

        solver = MIPSolver(self._tracking_config)

        print(f"Solving ILP batch {index}")
        print("Constructing ILP ...")

        self._add_nodes(solver=solver, index=index)
        self._add_edges(solver=solver, index=index)

        solver.set_standard_constraints()

        self._add_overlap_constraints(solver=solver, index=index)
        self._add_boundary_constraints(solver=solver, index=index)

        self._solver = solver

    @property
    def solver(self) -> MIPSolver:
        if self._solver is None:
            raise ValueError("`construct_model` must be called first.")
        return self._solver

    def __call__(
        self,
        index: int,
        use_annotations: bool = False,
        use_ground_truth_match: bool = False,
    ) -> None:
        """Queries data from the given index, add it to the solver, run it and push to the database.

        Parameters
        ----------
        index : int
            Batch index.
        use_annotations : bool
            Fix ILP variables using annotations, by default False.
        use_ground_truth_match : bool
            Fix ILP variables using ground truth matching data, by default False.
        """
        self.construct_model(index)

        if use_annotations and use_ground_truth_match:
            raise ValueError(
                "Only one of `use_annotations` and `use_ground_truth_match` can be True."
            )

        if use_annotations:
            print("Fixing annotations ...")
            self.fix_annotations(index)
        elif use_ground_truth_match:
            print("Fixing ground truth matches ...")
            self.fix_ground_truth_matches(index)

        print("Solving ILP ...")
        self.solve()

        print("Saving solution ...")
        self.add_solution(index)

        print("Done!")

    def fix_annotations(self, index: int) -> None:
        """
        Fix ILP variables using annotations.

        Parameters
        ----------
        index : int
            Batch index.
        """

        engine = sqla.create_engine(self._data_config.database_path)

        start_time, end_time = self._window_limits(index, True)

        with Session(engine) as session:
            # setting extra slack variables
            for mode, value in itertools.product(
                ["appear", "disappear", "division", "node"],
                [True, False],
            ):
                column = getattr(NodeDB, f"{mode}_annot")
                enum_value = VarAnnotation.REAL if value else VarAnnotation.FAKE
                indices = [
                    i
                    for i, in session.query(NodeDB.id).where(
                        NodeDB.t.between(start_time, end_time), column == enum_value
                    )
                ]
                self.solver.enforce_nodes_solution_value(indices, mode, value)

            # setting links
            for value in (True, False):
                enum_value = VarAnnotation.REAL if value else VarAnnotation.FAKE
                query = (
                    session.query(LinkDB.source_id, LinkDB.target_id)
                    .join(NodeDB, NodeDB.id == LinkDB.source_id)
                    .where(
                        NodeDB.t.between(start_time, end_time),
                        LinkDB.annotation == enum_value,
                    )
                )
                df = pd.read_sql(query.statement, session.bind)
                self.solver.enforce_edges_solution_value(
                    df["source_id"],
                    df["target_id"],
                    value,
                )

    def fix_ground_truth_matches(self, index: int) -> None:
        """
        Fix ILP variables using ground truth matching data.

        Parameters
        ----------
        index : int
            Batch index.
        """
        engine = sqla.create_engine(self._data_config.database_path)

        start_time, end_time = self._window_limits(index, True)

        with Session(engine) as session:
            # setting extra slack variables
            indices = [
                i
                for i, in session.query(GTLinkDB.source_id)  # equivalent to NodeDB.id
                .join(NodeDB, NodeDB.id == GTLinkDB.source_id)
                .where(NodeDB.t.between(start_time, end_time), GTLinkDB.selected)
            ]
            self.solver.enforce_nodes_solution_value(indices, "node", True)

    def solve(self) -> None:
        """Tracks by solving optimization model."""
        self.solver.optimize()

    def set_number_of_segments(self, time: int, number_of_segments: int) -> None:
        """Sets as a constraint the number of segments for a given time point.

        Parameters
        ----------
        time : int
            Time point.
        number_of_segments : int
            Number of segments.

        Examples
        --------
        >>> # creating tracking problem
        >>> segment(foreground, edges, config)
        >>> link(config)
        >>>
        >>> # low level API tracking
        >>> tracker = SQLTracking(config)
        >>> tracker.construct_model()
        >>> tracker.set_number_of_segments(0, 2)
        >>> tracker.solve()
        >>> tracker.add_solution()
        """
        engine = sqla.create_engine(self._data_config.database_path)
        with Session(engine) as session:
            indices = [i for i, in session.query(NodeDB.id).where(NodeDB.t == time)]

        self.solver.set_nodes_sum(indices, number_of_segments)

    def _window_limits(self, index: int, with_overlap: bool) -> Tuple[int, int]:
        """Computes time window of a given index, with or without overlap.

        Parameters
        ----------
        index : int
            Batch index.
        with_overlap : bool
            Flag indicating if `overlap_size` should be taken into account or not.

        Returns
        -------
        Tuple[int, int]
            Lower and upper window boundary.
        """
        start_time = (
            index * self._window_size
            - with_overlap * self._tracking_config.overlap_size
        )
        end_time = (
            index + 1
        ) * self._window_size + with_overlap * self._tracking_config.overlap_size
        return start_time, end_time - 1

    def _add_nodes(self, solver: BaseSolver, index: int) -> None:
        """Query nodes from a given batch index and add them to solver.

        Parameters
        ----------
        index : int
            Batch index.
        """
        start_time, end_time = self._window_limits(index, True)

        engine = sqla.create_engine(self._data_config.database_path)
        border_distance = self._tracking_config.image_border_size
        shape = self._data_config.metadata["shape"]
        data_dim = len(shape)

        if border_distance is None or np.all(border_distance == 0):
            with Session(engine) as session:
                query = session.query(
                    NodeDB.id,
                    NodeDB.t,
                    NodeDB.node_prob,
                ).where(NodeDB.t.between(start_time, end_time))
                df = pd.read_sql(query.statement, session.bind)
            is_border = False
        else:
            # Handle border distance
            # Ensure border_distance and shape have the same dimension
            if len(border_distance) < 3:
                border_distance = (0,) * (3 - len(border_distance)) + border_distance

            if data_dim == 3:
                # TYX -> TZYX
                shape = (shape[-3], 1, shape[-2], shape[-1])

            with Session(engine) as session:
                query = session.query(
                    NodeDB.id,
                    NodeDB.t,
                    NodeDB.z,
                    NodeDB.y,
                    NodeDB.x,
                    NodeDB.node_prob,
                ).where(NodeDB.t.between(start_time, end_time))
                df = pd.read_sql(query.statement, session.bind)

            is_border = _check_inside_border(df, border_distance, shape)

        n_invalid_prob = (df["node_prob"] < 0).sum()

        if n_invalid_prob == df.shape[0]:
            nodes_prob = None
        elif n_invalid_prob == 0:
            nodes_prob = df["node_prob"]
        else:
            raise ValueError(
                "None or all nodes' probabilities must be provided found "
                f"Found {df.shape[0] - n_invalid_prob} / {df.shape[0]} valid probs."
            )

        start_time = max(start_time, 0)
        end_time = min(end_time, self._max_t)

        LOG.info(f"Batch {index}, nodes with t between {start_time} and {end_time}")

        solver.add_nodes(
            df["id"],
            df["t"] == start_time,
            df["t"] == end_time,
            is_border=is_border,
            nodes_prob=nodes_prob,
        )

    def _add_edges(self, solver: BaseSolver, index: int) -> None:
        """Query edges from a given batch index and add them to solver.

        Parameters
        ----------
        index : int
            Batch index.
        """
        start_time, end_time = self._window_limits(index, True)

        engine = sqla.create_engine(self._data_config.database_path)
        with Session(engine) as session:
            query = (
                session.query(LinkDB)
                .join(NodeDB, NodeDB.id == LinkDB.source_id)
                .where(NodeDB.t.between(start_time, end_time - 1))
                # subtracting one because we're using source_id as reference
            )
            df = pd.read_sql(query.statement, session.bind)

        LOG.info(
            f"Batch {index}, edges with source nodes with t between {start_time} and {end_time - 1}"
        )

        solver.add_edges(df["source_id"], df["target_id"], df["weight"])

    def _add_overlap_constraints(self, solver: BaseSolver, index: int) -> None:
        """Adds overlaping segmentation constrainsts

        Parameters
        ----------
        index : int
            Batch index.
        """
        start_time, end_time = self._window_limits(index, True)

        engine = sqla.create_engine(self._data_config.database_path)
        with Session(engine) as session:
            query = (
                session.query(OverlapDB)
                .join(NodeDB, NodeDB.id == OverlapDB.node_id)
                .where(NodeDB.t.between(start_time, end_time))
            )
            df = pd.read_sql(query.statement, session.bind)

        solver.add_overlap_constraints(df["node_id"], df["ancestor_id"])

    def _add_boundary_constraints(self, solver: BaseSolver, index: int) -> None:
        """
        Enforce to solution nodes from the boundary (in time) already selected from adjacent batches.

        Parameters
        ----------
        index : int
            Batch index.
        """
        start_time, end_time = self._window_limits(index, True)

        engine = sqla.create_engine(self._data_config.database_path)
        with Session(engine) as session:
            query = session.query(NodeDB.id).where(NodeDB.selected)

            start_nodes = [n for n, in query.where(NodeDB.t == start_time)]
            end_nodes = [n for n, in query.where(NodeDB.t == end_time)]

        LOG.info(
            f"# {len(start_nodes)} boundary constraints found at at t = {start_time}"
        )
        LOG.info(f"# {len(end_nodes)} boundary constraints found at at t = {end_time}")

        solver.enforce_nodes_solution_value(start_nodes, variable="node", value=True)
        solver.enforce_nodes_solution_value(end_nodes, variable="node", value=True)

    def add_solution(self, index: int = 0) -> None:
        """Adds selected nodes to solution in database.

        Parameters
        ----------
        index : int
            Batch index, by default 0, which works for single batch tracking.
        """
        solution = self.solver.solution()

        solution["node_id"] = solution.index

        start_time, end_time = self._window_limits(index, False)

        engine = sqla.create_engine(self._data_config.database_path)
        with Session(engine) as session:
            general_stmt = (
                sqla.update(NodeDB)
                .where(
                    NodeDB.t.between(start_time, end_time),
                    NodeDB.id == sqla.bindparam("node_id"),
                )
                .values(parent_id=sqla.bindparam("parent_id"), selected=True)
            )
            session.connection().execute(
                general_stmt,
                solution[["node_id", "parent_id"]].to_dict("records"),
                execution_options={"synchronize_session": False},
            )

            # condition isn't necessary but avoids a useless operation
            if start_time > 0:
                # insert nodes from start time - 1 without their parent
                start_stmt = (
                    sqla.update(NodeDB)
                    .where(
                        NodeDB.t == start_time - 1,
                        NodeDB.id == sqla.bindparam("node_id"),
                    )
                    .values(selected=True)
                )
                session.connection().execute(
                    start_stmt,
                    solution[["node_id"]].to_dict("records"),
                    execution_options={"syncronize_session": False},
                )

            session.commit()

    def reset_solution(self) -> None:
        """Resets every node parent_id and selected variables."""
        self.clear_solution_from_database(self._data_config.database_path)

    @staticmethod
    def clear_solution_from_database(database_path: str) -> None:
        LOG.info("Clearing nodes database solutions.")
        engine = sqla.create_engine(database_path)
        with Session(engine) as session:
            statement = sqla.update(NodeDB).values(parent_id=NO_PARENT, selected=False)
            session.execute(statement)
            session.commit()


def _check_inside_border(
    df: pd.DataFrame,
    border_distance: Tuple[int, int, int],
    shape: Tuple[int, int, int, int],
) -> np.ndarray:
    """
    Check if nodes are inside the border.

    Parameters
    ----------
    df : pd.DataFrame
        Nodes dataframe.
    border_distance : Tuple[int, int, int]
        Border distance in pixels (Z, Y, X).
    shape : Tuple[int, int, int, int]
        Image shape (T, Z, Y, X).

    Returns
    -------
    np.ndarray
        Boolean array indicating if nodes are inside the border.
    """
    inside_border = np.logical_and.reduce(
        [
            (border_distance[i] <= df[c]) & (df[c] <= shape[i] - border_distance[i])
            for i, c in zip([-3, -2, -1], ["z", "y", "x"])
        ]
    )
    is_border = ~inside_border
    # Log the coordinates of nodes inside the border if logging level is debug
    if LOG.isEnabledFor(logging.DEBUG):
        for row in df[is_border].itertuples(index=False):
            LOG.debug(
                "Node with coords (z,y,x) at: (%s, %s, %s) is inside border.",
                row.z,
                row.y,
                row.x,
            )
    return is_border



import logging

import numpy as np
import pandas as pd
import sqlalchemy as sqla
from sqlalchemy.orm import Session

from ultrack.config import MainConfig
from ultrack.core.database import LinkDB
from ultrack.core.segmentation.processing import get_nodes_features

LOG = logging.getLogger(__name__)


def get_links_features(
    config: MainConfig,
    **kwargs,
) -> pd.DataFrame:
    """
    Creates a pandas dataframe from links features defined during linking.

    Parameters
    ----------
    config : MainConfig
        Configuration parameters.
    **kwargs : dict
        Keyword arguments passed to `get_nodes_features`.

    Returns
    -------
    pd.DataFrame:
        Links dataframe with the features.
        MultiIndex columns:
        * target_id: target node id
        * source_id: source node id
    """
    nodes_features = get_nodes_features(config, **kwargs)

    engine = sqla.create_engine(config.data_config.database_path)

    with Session(engine) as session:
        query = session.query(
            LinkDB.source_id,
            LinkDB.target_id,
        )
        links_df = pd.read_sql(query.statement, session.bind)
        LOG.info("Loaded %d links.", len(links_df))

    source_df = nodes_features.loc[links_df["source_id"]]
    target_df = nodes_features.loc[links_df["target_id"]]

    LOG.info("Source features shape: %s", source_df.shape)
    LOG.info("Target features shape: %s", target_df.shape)

    if "x" in source_df.columns:
        # compute distance between source and target
        links_df["dist"] = np.sqrt(
            np.sum(
                [
                    np.square(source_df[c].to_numpy() - target_df[c].to_numpy())
                    for c in ["x", "y", "z"]
                ],
                axis=0,
            )
        )

    cols = source_df.columns.drop(["id", "t", "z", "y", "x"], errors="ignore")

    links_df[cols] = source_df[cols].to_numpy() - target_df[cols].to_numpy()

    links_df.set_index(["target_id", "source_id"], inplace=True)

    if links_df.empty:
        raise ValueError("Dataset is empty after computing the features")

    return links_df



import numpy as np
import pytest

from ultrack.core.solve.solver.heuristic._heap import Heap, Policy


@pytest.mark.parametrize(
    "policy",
    [Policy.minimum, Policy.maximum],
)
def test_heap(policy: Policy) -> None:
    size = 10_000
    rng = np.random.default_rng(42)

    values = rng.uniform(size=size).astype(np.float32)
    heap = Heap(values, policy=policy)

    if policy == Policy.minimum:
        nil_value = np.finfo(np.float32).max
        arg_func = np.argmin
    else:
        nil_value = np.finfo(np.float32).min
        arg_func = np.argmax

    count = 0
    # testing partial insertion with removal
    for i in range(size):
        heap.insert(i)
        if (i + 1) % 5 == 0:
            j = heap.pop()
            assert j == arg_func(values[: i + 1])
            values[j] = nil_value
            count += 1

    # testing removal until it's empty.
    while not heap.is_empty():
        i = heap.pop()
        assert i == arg_func(values)
        values[i] = nil_value
        count += 1

    # asserting the right number of elements where removed
    assert count == size


def test_insert_array() -> None:
    size = 1_000
    rng = np.random.default_rng(42)

    values = rng.uniform(size=size).astype(np.float32)
    heap = Heap(values, policy=Policy.minimum)

    heap.insert_array(np.arange(size))
    while not heap.is_empty():
        i = heap.pop()
        assert i == np.argmin(values)
        values[i] = np.finfo(np.float32).max



import logging
from typing import Optional

from ultrack.config.config import MainConfig
from ultrack.core.solve.sqltracking import SQLTracking

LOG = logging.getLogger(__name__)


def solve(
    config: MainConfig,
    batch_index: Optional[int] = None,
    overwrite: bool = False,
    use_annotations: bool = False,
    use_ground_truth_match: bool = False,
) -> None:
    """Compute tracking by selecting nodes with maximum flow from database.

    Parameters
    ----------
    config : MainConfig
        Configuration parameters.
    batch_index : Optional[int], optional
        Batch index for processing a subset of nodes, by default everything is processed.
    overwrite : bool, optional
        Resets existing solution before processing.
    use_annotations : bool, optional
        Use annotations to fix ILP variables, by default False
    use_ground_truth_match : bool
        Fix ILP variables using ground truth matching data, by default False.
    """
    tracker = SQLTracking(config)

    if overwrite and (batch_index is None or batch_index == 0):
        tracker.reset_solution()

    if use_annotations or use_ground_truth_match:
        if not overwrite:
            LOG.warning(
                "Annotations or ground-truth matches are being used without overwriting "
                "the existing solution.\nIgnore this warning if you this is your first tracking call."
            )

    kwargs = {
        "use_annotations": use_annotations,
        "use_ground_truth_match": use_ground_truth_match,
    }

    if isinstance(batch_index, int):
        LOG.info(f"Solving ILP of batch index {batch_index}")
        tracker(index=batch_index, **kwargs)
    else:
        LOG.info(f"Solving ILP batch indices from 0 to {tracker.num_batches}")
        # interleaved processing
        for batch_index in range(0, tracker.num_batches, 2):
            tracker(index=batch_index, **kwargs)
        for batch_index in range(1, tracker.num_batches, 2):
            tracker(index=batch_index, **kwargs)
